#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <map>
#include <string>

#include "Common.h"
#include "OpCodes.h"
#include "Class.h"
#include "Heap.h"
#include "Stack.h"
#include "JVM.h"

//-----------------------------------------------------------------------------
// Name: cJavaClass::cJavaClass()
//-----------------------------------------------------------------------------
cJavaClass::cJavaClass(cJVM *pOwner)
{
	m_ppConstantPool = NULL;
	m_pMethodInfos = NULL;
	m_pMainMethod = NULL;
	m_pInterfaces = NULL;
	m_pClassInit = NULL;
	m_pSuperClass = NULL;
	m_pOwner = pOwner;
	m_strThisClass = NULL;
	m_iSize = 0;
	m_pFieldInfos = NULL;
	m_pStaticData = NULL;
	m_bInitialized = false;

	m_iMagic = 0;
	m_iMinor = 0;
	m_iMajor = 0;
	m_iConstantPoolCount = 0;
	m_iAccessFlags = 0;
	m_iThisClass = 0;
	m_iSuperClass = 0;
	m_iInterfaceCount = 0;
	m_iFieldsCount = 0;
	m_iMethodsCount = 0;
	m_iSize = 0;
	m_iStaticSize = 0;
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::~cJavaClass()
//-----------------------------------------------------------------------------
cJavaClass::~cJavaClass()
{
	if(m_ppConstantPool != NULL)
	{
		for(int i = 0;i < m_iConstantPoolCount;++i)
		{
			if(m_ppConstantPool[i] != NULL)
			{
				delete m_ppConstantPool[i];
			}
		}

		delete[] m_ppConstantPool;
	}

	SAFE_DELETE_ARRAY(m_pMethodInfos);
	SAFE_DELETE_ARRAY(m_pFieldInfos);
	SAFE_DELETE_ARRAY(m_pStaticData);
	SAFE_DELETE_ARRAY(m_pInterfaces);
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::GetClassFromPool()
//-----------------------------------------------------------------------------
cJavaClass* cJavaClass::GetClassFromPool(u2 iIndex) 
{
	return ((CONSTANT_Class*)m_ppConstantPool[iIndex])->pClass; 
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::GetMethodFromPool()
//-----------------------------------------------------------------------------
METHOD_INFO* cJavaClass::GetMethodFromPool(u2 iIndex) 
{
	return ((CONSTANT_MethodRef*)m_ppConstantPool[iIndex])->pMethod; 
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::GetInterfaceMethodFromPool()
//-----------------------------------------------------------------------------
METHOD_INFO* cJavaClass::GetInterfaceMethodFromPool(u2 iIndex) 
{
	return ((CONSTANT_InterfaceMethodRef*)m_ppConstantPool[iIndex])->pMethod; 
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::GetFieldFromPool()
//-----------------------------------------------------------------------------
FIELD_INFO* cJavaClass::GetFieldFromPool(u2 iIndex)
{
	return ((CONSTANT_FieldRef*)m_ppConstantPool[iIndex])->pField;
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::GetInterfaceMethodNameAndDesc()
//-----------------------------------------------------------------------------
void cJavaClass::GetInterfaceMethodNameAndDesc(u2 iIndex, u1 **ppName, u1 **ppDesc)
{
	CONSTANT_InterfaceMethodRef *pRef = (CONSTANT_InterfaceMethodRef*)m_ppConstantPool[iIndex];
	CONSTANT_NameAndType *pType = (CONSTANT_NameAndType*)m_ppConstantPool[pRef->iNameAndTypeIndex];
	CONSTANT_Utf8 *pName = (CONSTANT_Utf8*)m_ppConstantPool[pType->iNameIndex];
	CONSTANT_Utf8 *pDesc = (CONSTANT_Utf8*)m_ppConstantPool[pType->iDescriptorIndex];

	*ppName = pName->pData;
	*ppDesc = pDesc->pData;
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::SetStaticValue()
//-----------------------------------------------------------------------------
void cJavaClass::SetStaticValue(u2 iOffset, void *pData, u2 iSize)
{
	int i = *(int*)pData;
	//int *p = (int*)&m_pStaticData[iOffset];
	//memcpy(&m_pStaticData[iOffset],&i,iSize);
	memcpy(&m_pStaticData[iOffset],pData,iSize);
	//memcpy(p,&i,iSize);
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::GetStaticValue()
//-----------------------------------------------------------------------------
void cJavaClass::GetStaticValue(u2 iOffset, void *pData, u2 iSize)
{
	//OBJECT *p;
	memcpy(pData,&m_pStaticData[iOffset],iSize);
	//p = (OBJECT *)*pData;
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::GetDoubleOrLongFromConstantPool()
//-----------------------------------------------------------------------------
bool cJavaClass::GetDoubleOrLongFromConstantPool(u2 iIndex,u4 &iHighBytes, u4 &iLowBytes)
{
	u1 tag = *(u1*)m_ppConstantPool[iIndex];
	//make sure it is double or long
	if(tag != 5 && tag != 6)
		return false;

	CONSTANT_Long *pConst = (CONSTANT_Long*)m_ppConstantPool[iIndex];
	iHighBytes = pConst->iHighBytes;
	iLowBytes = pConst->iLowBytes;
	return true;
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::GetFloatOrIntFromConstantPool()
//-----------------------------------------------------------------------------
bool cJavaClass::GetFloatOrIntFromConstantPool(u2 iIndex,u4 *pData)
{
	u1 tag = *(u1*)m_ppConstantPool[iIndex];

	//int
	if(tag == 3)
	{
		CONSTANT_Integer *pConst = (CONSTANT_Integer*)m_ppConstantPool[iIndex];
		*pData = pConst->iInteger;
		return true;
	}
	//float
	if(tag == 4)
	{
		CONSTANT_Float *pConst = (CONSTANT_Float*)m_ppConstantPool[iIndex];
		memcpy(pData,&pConst->fFloat,sizeof(jFloat));
		return true;
	}

	return false;
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::LoadConstantPool()
//-----------------------------------------------------------------------------
void cJavaClass::Debug_PrintConstantPool()
{
	for(int i = 1;i < this->m_iConstantPoolCount;++i)
	{
		if(m_ppConstantPool[i] == NULL)
		{
			printf("%d: NULL\n",i);
			continue;
		}
		u1 tag = *(u1*)m_ppConstantPool[i];
		
		printf("%d: ",i);
		switch(tag)
		{
		case 7: //class
			((CONSTANT_Class*)m_ppConstantPool[i])->PrintInfo();
			break;

		case 9: //field ref
			((CONSTANT_FieldRef*)m_ppConstantPool[i])->PrintInfo();
			break;

		case 10: //method ref
			((CONSTANT_MethodRef*)m_ppConstantPool[i])->PrintInfo();
			break;

		case 11: //interface method ref
			((CONSTANT_InterfaceMethodRef*)m_ppConstantPool[i])->PrintInfo();
			break;

		case 8: //string
			((CONSTANT_String*)m_ppConstantPool[i])->PrintInfo();
			break;

		case 3: //integer
			((CONSTANT_Integer*)m_ppConstantPool[i])->PrintInfo();
			break;

		case 4: //float
			((CONSTANT_Float*)m_ppConstantPool[i])->PrintInfo();
			break;

		case 12: //name and type
			((CONSTANT_NameAndType*)m_ppConstantPool[i])->PrintInfo();
			break;

		case 1: //utf 8
			((CONSTANT_Utf8*)m_ppConstantPool[i])->PrintInfo();			
			break;

		case 5: //long			
			((CONSTANT_Long*)m_ppConstantPool[i])->PrintInfo();
			break;

		case 6: //double
			((CONSTANT_Double*)m_ppConstantPool[i])->PrintInfo();
			break;
		}
	}
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::LoadConstantPool()
//-----------------------------------------------------------------------------
void cJavaClass::LoadConstantPool(FILE *f)
{
	m_iConstantPoolCount = ReadU2(f);
	if(m_iConstantPoolCount == 0)
		return;

	m_ppConstantPool = new void*[m_iConstantPoolCount];

	//first index is not used
	m_ppConstantPool[0] = NULL;

	for(int i = 1;i < m_iConstantPoolCount;++i)
	{
		u1 tag = ReadU1(f);
		u2 h1,h2;
		u4 d1,d2 = 0;
		switch(tag)
		{
		case 7: //class
			m_ppConstantPool[i] = new CONSTANT_Class(tag,ReadU2(f));
			break;

		case 9: //field ref
			h1 = ReadU2(f);
			h2 = ReadU2(f);
			m_ppConstantPool[i] = new CONSTANT_FieldRef(tag,h1,h2);
			break;

		case 10: //method ref
			h1 = ReadU2(f);
			h2 = ReadU2(f);
			m_ppConstantPool[i] = new CONSTANT_MethodRef(tag,h1,h2);
			break;

		case 11: //interface method ref
			h1 = ReadU2(f);
			h2 = ReadU2(f);
			m_ppConstantPool[i] = new CONSTANT_InterfaceMethodRef(tag,h1,h2);
			break;

		case 8: //string
			m_ppConstantPool[i] = new CONSTANT_String(tag,ReadU2(f));
			break;

		case 3: //integer
			m_ppConstantPool[i] = new CONSTANT_Integer(tag,ReadU4(f));
			break;

		case 4: //float
			m_ppConstantPool[i] = new CONSTANT_Float(tag,ReadU4(f));
			break;

		case 12: //name and type
			h1 = ReadU2(f);
			h2 = ReadU2(f);
			m_ppConstantPool[i] = new CONSTANT_NameAndType(tag,h1,h2);
			break;

		case 5: //long
			d1 = ReadU4(f);
			d2 = ReadU4(f);
			m_ppConstantPool[i] = new CONSTANT_Long(tag,d1,d2);
			++i;
			m_ppConstantPool[i] = NULL;
			break;

		case 6: //double
			d1 = ReadU4(f);
			d2 = ReadU4(f);
			m_ppConstantPool[i] = new CONSTANT_Double(tag,d1,d2);
			++i;
			m_ppConstantPool[i] = NULL;
			break;

		case 1: //utf 8 - it isn't really utf8, it is simple 1 byte char
			m_ppConstantPool[i] = new CONSTANT_Utf8(tag,ReadU2(f));
			//allocate string
			((CONSTANT_Utf8*)m_ppConstantPool[i])->pData = new u1[((CONSTANT_Utf8*)m_ppConstantPool[i])->iLength + 1];
			//read string from file
			fread(((CONSTANT_Utf8*)m_ppConstantPool[i])->pData,sizeof(u1),((CONSTANT_Utf8*)m_ppConstantPool[i])->iLength,f);
			//zero terminate string
			((CONSTANT_Utf8*)m_ppConstantPool[i])->pData[((CONSTANT_Utf8*)m_ppConstantPool[i])->iLength] = 0;
			break;

		default:
			m_ppConstantPool[i] = NULL;
			break;
		};		
	}
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::LoadInterfaces()
//-----------------------------------------------------------------------------
void cJavaClass::LoadInterfaces(FILE *f)
{
	//num interfaces
	m_iInterfaceCount = ReadU2(f);

	if(m_iInterfaceCount == 0)
		return;

	m_pInterfaces = new u2[m_iInterfaceCount];
	for(int i = 0;i < m_iInterfaceCount;++i)
	{
		m_pInterfaces[i] = ReadU2(f);
	}
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::LoadFieldInfos()
//-----------------------------------------------------------------------------
void cJavaClass::LoadFieldInfos(FILE *f)
{
	//num fields
	this->m_iFieldsCount = ReadU2(f);

	if(m_iFieldsCount == 0)
		return;

	m_pFieldInfos = new FIELD_INFO[m_iFieldsCount];

	//for each field
	for(int i = 0;i < m_iFieldsCount;++i)
	{
		m_pFieldInfos[i].pOwner = this;

		m_pFieldInfos[i].iAccessFlags = ReadU2(f);
		m_pFieldInfos[i].iNameIndex = ReadU2(f);
		m_pFieldInfos[i].iDescriptorIndex = ReadU2(f);

		m_pFieldInfos[i].iAttributesCount = ReadU2(f);

		if(m_pFieldInfos[i].iAttributesCount == 0)
			continue;

		m_pFieldInfos[i].pAttributeInfos = new ATTRIBUTE_INFO[m_pFieldInfos[i].iAttributesCount];
		for(int j = 0;j < m_pFieldInfos[i].iAttributesCount;++j)
		{
			m_pFieldInfos[i].pAttributeInfos[j].iNameIndex = ReadU2(f);
			m_pFieldInfos[i].pAttributeInfos[j].iLength = ReadU4(f);
			m_pFieldInfos[i].pAttributeInfos[j].pInfo = new u1[m_pFieldInfos[i].pAttributeInfos[j].iLength];
			fread(m_pFieldInfos[i].pAttributeInfos[j].pInfo,sizeof(u1),m_pFieldInfos[i].pAttributeInfos[j].iLength,f);

			//if attribute is constant value, keep constant index for future reference
			u1 *strName = ((CONSTANT_Utf8*)m_ppConstantPool[m_pFieldInfos[i].pAttributeInfos[j].iNameIndex])->pData;
			if(0 == strcmp((const char*)strName,"ConstantValue"))
			{
				u1 v1 = *m_pFieldInfos[i].pAttributeInfos[j].pInfo;
				u1 v2 = *(m_pFieldInfos[i].pAttributeInfos[j].pInfo + 1);
				m_pFieldInfos[i].iConstantValueIndex = (u2)((v1 << 8) | v2);
			}
		}
	}
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::LoadMethodInfos()
//-----------------------------------------------------------------------------
void cJavaClass::LoadMethodInfos(FILE *f)
{
	//num methods
	m_iMethodsCount = ReadU2(f);

	if(m_iMethodsCount == 0)
		return;

	m_pMethodInfos = new METHOD_INFO[m_iMethodsCount];

	//for each method
	for(int i = 0;i < m_iMethodsCount;++i)
	{
		//set owner
		m_pMethodInfos[i].pOwner = this;

		m_pMethodInfos[i].iAccessFlags = ReadU2(f);
		m_pMethodInfos[i].iNameIndex = ReadU2(f);

		//check for main method
		u1 *strName = ((CONSTANT_Utf8*)m_ppConstantPool[m_pMethodInfos[i].iNameIndex])->pData;
		if(strcmp((const char*)strName, "main") == 0)
		{
			this->m_pMainMethod = &m_pMethodInfos[i];
		}

		//check for class initialization method
		if(strcmp((const char*)strName, "<clinit>") == 0)
		{
			this->m_pClassInit = &m_pMethodInfos[i];
		}

		//get method's descriptor
		m_pMethodInfos[i].iDescriptorIndex = ReadU2(f);
		u1 *strDescriptor = ((CONSTANT_Utf8*)m_ppConstantPool[m_pMethodInfos[i].iDescriptorIndex])->pData;

		//parse method descriptor for arguments size
		m_pMethodInfos[i].iNumArguments = ParseAgrumentCount(strDescriptor);

		//read attributes
		m_pMethodInfos[i].iAttrbutesCount = ReadU2(f);

		m_pMethodInfos[i].pAttributeInfos = new ATTRIBUTE_INFO[m_pMethodInfos[i].iAttrbutesCount];

		for(int j = 0;j < m_pMethodInfos[i].iAttrbutesCount;++j)
		{
			u2 nameIndex = ReadU2(f);
			u1 *pName = ((CONSTANT_Utf8*)m_ppConstantPool[nameIndex])->pData;
			//check for code attribute (bytecode)
			if(0 == strcmp((const char*)pName,"Code"))
			{
				//read code attribute
				m_pMethodInfos[i].pAttributeInfos[j].iNameIndex = nameIndex;
				m_pMethodInfos[i].pAttributeInfos[j].iLength = ReadU4(f);
				m_pMethodInfos[i].pAttributeInfos[j].pInfo = NULL;

				m_pMethodInfos[i].iMaxStack = ReadU2(f);
				m_pMethodInfos[i].iMaxLocals = ReadU2(f);

				m_pMethodInfos[i].iCodeLength = ReadU4(f);
				//read bytecode
				m_pMethodInfos[i].pCode = new u1[m_pMethodInfos[i].iCodeLength];
				fread(m_pMethodInfos[i].pCode,sizeof(u1),m_pMethodInfos[i].iCodeLength,f);

				m_pMethodInfos[i].iExceptionTableLength = ReadU2(f);

				//read exception table
				if(m_pMethodInfos[i].iExceptionTableLength != 0)
				{
					m_pMethodInfos[i].pExceptionTable = new EXCEPTION_INFO[m_pMethodInfos[i].iExceptionTableLength];

					for(int k = 0;k < m_pMethodInfos[i].iExceptionTableLength;++k)
					{
						m_pMethodInfos[i].pExceptionTable[k].iStartPC = ReadU2(f);
						m_pMethodInfos[i].pExceptionTable[k].iEndPC = ReadU2(f);
						m_pMethodInfos[i].pExceptionTable[k].iHandlerPC = ReadU2(f);
						m_pMethodInfos[i].pExceptionTable[k].iCatchType = ReadU2(f);
					}
				}

				//read code attributes if any
				m_pMethodInfos[i].iCodeAttributesCount = ReadU2(f);
				if(m_pMethodInfos[i].iCodeAttributesCount != 0)
				{
					m_pMethodInfos[i].pCodeAttributeInfos = new ATTRIBUTE_INFO[m_pMethodInfos[i].iCodeAttributesCount];
					for(int k = 0;k < m_pMethodInfos[i].iCodeAttributesCount;++k)
					{
						m_pMethodInfos[i].pCodeAttributeInfos[k].iNameIndex = ReadU2(f);
						m_pMethodInfos[i].pCodeAttributeInfos[k].iLength = ReadU4(f);
						m_pMethodInfos[i].pCodeAttributeInfos[k].pInfo = new u1[m_pMethodInfos[i].pCodeAttributeInfos[k].iLength];
						fread(m_pMethodInfos[i].pCodeAttributeInfos[k].pInfo,sizeof(u1),m_pMethodInfos[i].pCodeAttributeInfos[k].iLength,f);
					}
				}
			}
			else
			{
				//read other method's atributes
				m_pMethodInfos[i].pAttributeInfos[j].iNameIndex = nameIndex;
				m_pMethodInfos[i].pAttributeInfos[j].iLength = ReadU4(f);
				m_pMethodInfos[i].pAttributeInfos[j].pInfo = new u1[m_pMethodInfos[i].pAttributeInfos[j].iLength];
				fread(m_pMethodInfos[i].pAttributeInfos[j].pInfo,sizeof(u1),m_pMethodInfos[i].pAttributeInfos[j].iLength,f);
			}
		}
	}
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::ParseAgrumentCount()
//-----------------------------------------------------------------------------
u2 cJavaClass::ParseAgrumentCount(u1 *strDescriptor)
{
	u2 iNumArgs = 0;

	int iLength = strlen((const char*)strDescriptor);
	bool bInArgs = false;
	bool bInClassname = false;

	u2 iDoubleSize = 2;
	//go through methods descriptor
	//double and long arguments take up 2 stack values
	//every other data type takes up 1 stack value
	for(int i = 0;i < iLength;++i)
	{
		if(bInClassname)
		{
			if(strDescriptor[i] == ';')
				bInClassname = false;
		}

		if(strDescriptor[i] == '(')
		{
			bInArgs = true;
			continue;
		}
		if(strDescriptor[i] == ')')
		{
			bInArgs = false;
			continue;
		}

		if(strDescriptor[i] == '[')
		{
			iDoubleSize = 1;
			continue;
		}

		if(bInArgs)
		{
			switch(strDescriptor[i])
			{
			case 'B'://byte
			case 'C'://char
			case 'F'://float
			case 'I'://int
			case 'L'://reference
			case 'S'://short
			case 'Z'://bool
				++iNumArgs;
				break;

			case 'D'://double
			case 'J'://long
				iNumArgs += iDoubleSize;
				iDoubleSize = 2;
				break;
			};
		}
	}

	return iNumArgs;
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::ParseObjectName()
//-----------------------------------------------------------------------------
u1* cJavaClass::ParseObjectName(u1 *name)
{
	//name could be class, interface or (multi dimensional)array
	//we want only class, interface name
	while(*name == '[')
	{
		++name;
	}

	if(strlen((char*)name) == 1)
	{
		switch(*name)
		{
		case 'B'://byte
		case 'C'://char
		case 'F'://float
		case 'I'://int
		case 'S'://short
		case 'Z'://bool
		case 'D'://double
		case 'J'://long
			return 0;
		};
	}

	return name;
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::ParseArrayDefinition()
//-----------------------------------------------------------------------------
cJavaClass* cJavaClass::ParseArrayDefinition(u2 iConstantIndex, u1 &iDimensions, u1 &iArrayItemSize)
{
	CONSTANT_Class *pClass;
	u1 *pName;

	pClass = (CONSTANT_Class*)m_ppConstantPool[iConstantIndex];
	pName = ((CONSTANT_Utf8*)m_ppConstantPool[pClass->iIndex])->pData;

	iDimensions = 0;
	while(*pName == '[')
	{
		++iDimensions;
		++pName;
	}

	if(strlen((char*)pName) == 1)
	{
		switch(*pName)
		{
		case 'B'://byte
		case 'Z'://bool
			iArrayItemSize = sizeof(jByte);
			return NULL;
		case 'C'://char
		case 'S'://short
			iArrayItemSize = sizeof(jChar);
			return NULL;
		case 'F'://float
		case 'I'://int
			iArrayItemSize = sizeof(jFloat);
			return NULL;
		case 'D'://double
		case 'J'://long
			iArrayItemSize = sizeof(jDouble);
			return NULL;
		};
	}

	//object reference
	iArrayItemSize = sizeof(void*);
	return pClass->pClass;
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::ResolveSymbolicReferences()
//-----------------------------------------------------------------------------
void cJavaClass::ResolveSymbolicReferences()
{
	cJavaClass* pJavaClass;
	std::string s;
	CONSTANT_Class *pClass;
	u1 *pName;
	
	if(m_iSuperClass == 0)
	{ //no super class
		m_pSuperClass = NULL;
	}
	else
	{ //super class exists
		//get its full name
		pClass = (CONSTANT_Class*)m_ppConstantPool[this->m_iSuperClass];
		pName = ((CONSTANT_Utf8*)m_ppConstantPool[pClass->iIndex])->pData;

		//use cJVM and try to load super class from file
		pJavaClass = m_pOwner->LoadClass((const char*)pName);
		if(pJavaClass == NULL)
		{
			s = "Dependent class \"";
			s += (char*)pName;
			s += "\" failed to load.";
			throw s;
		}
		m_pSuperClass = pJavaClass;
	}

	CONSTANT_FieldRef *pFieldRef;
	CONSTANT_MethodRef *pMethodRef;
	CONSTANT_InterfaceMethodRef *pInterfaceMethodRef;
	CONSTANT_NameAndType *pNameAndType;
	u1 *strName;
	u1 *strDescriptor;
	//go trough all constant pool entries
	for(int i = 1;i < this->m_iConstantPoolCount;++i)
	{
		//check constant's tag
		u1 tag = *(u1*)m_ppConstantPool[i];
		
		switch(tag)
		{
		case 7: //class
			pClass = (CONSTANT_Class*)m_ppConstantPool[i];

			if(i == m_iThisClass)
			{
				//if currnt class -> update reference and continue...
				pClass->pClass = this;
				break;
			}

			//resolve classe's full name and try to load it using cJVM
			pName = ((CONSTANT_Utf8*)m_ppConstantPool[pClass->iIndex])->pData;

			//could be array def. we want only class name
			pName = ParseObjectName(pName);
			if(pName == NULL)
				break;

			pJavaClass = m_pOwner->LoadClass((const char*)pName);
			if(pJavaClass == NULL)
			{
				s = "Dependent class \"";
				s += (char*)pName;
				s += "\" failed to load.";
				throw s;
			}
			//update reference
			pClass->pClass = pJavaClass;
			break;

		case 9: //field ref
			pFieldRef = (CONSTANT_FieldRef*)m_ppConstantPool[i];

			//if field belongs to this class simply update reference
			if(pFieldRef->iClassIndex == m_iThisClass)
				pJavaClass = this;
			else
			{
				//otherwise, get class full name and try to load it from file
				pClass = (CONSTANT_Class*)m_ppConstantPool[pFieldRef->iClassIndex];
				pName = ((CONSTANT_Utf8*)m_ppConstantPool[pClass->iIndex])->pData;

				//could be array def. we want only class name
				pName = ParseObjectName(pName);
				if(pName == NULL)
					break;

				pJavaClass = m_pOwner->LoadClass((const char*)pName);
				if(pJavaClass == NULL)
				{
					s = "Dependent class \"";
					s += (char*)pName;
					s += "\" failed to load.";
					throw s;
				}
			}

			//now, get field name and descriptor
			pNameAndType = (CONSTANT_NameAndType*)m_ppConstantPool[pFieldRef->iNameAndTypeIndex];
			strName = ((CONSTANT_Utf8*)m_ppConstantPool[pNameAndType->iNameIndex])->pData;
			strDescriptor = ((CONSTANT_Utf8*)m_ppConstantPool[pNameAndType->iDescriptorIndex])->pData;

			//search parent class for field and update reference
			pFieldRef->pField = pJavaClass->FindField(strName,strDescriptor);
			if(pFieldRef->pField == NULL)
			{
				s = "Can't find field \"";
				s += (char*)strName;
				s += "\" in class \"";
				s += pJavaClass->GetName();
				s += "\".";
			}
			break;

		case 10: //method ref
			pMethodRef = (CONSTANT_MethodRef*)m_ppConstantPool[i];

			//if method belongs to current class simply update reference
			if(pMethodRef->iClassIndex == m_iThisClass)
				pJavaClass = this;
			else
			{
				//or get class full name and try to load it from file
				pClass = (CONSTANT_Class*)m_ppConstantPool[pMethodRef->iClassIndex];
				pName = ((CONSTANT_Utf8*)m_ppConstantPool[pClass->iIndex])->pData;

				//could be array def. we want only class name
				pName = ParseObjectName(pName);
				if(pName == NULL)
					break;

				//update ref
				pJavaClass = m_pOwner->LoadClass((const char*)pName);
				if(pJavaClass == NULL)
				{
					s = "Dependent class \"";
					s += (char*)pName;
					s += "\" failed to load.";
					throw s;
				}
			}

			//now we have class pointer
			//get method's name and descriptor
			pNameAndType = (CONSTANT_NameAndType*)m_ppConstantPool[pMethodRef->iNameAndTypeIndex];
			strName = ((CONSTANT_Utf8*)m_ppConstantPool[pNameAndType->iNameIndex])->pData;
			strDescriptor = ((CONSTANT_Utf8*)m_ppConstantPool[pNameAndType->iDescriptorIndex])->pData;

			//search class for method and update reference
			pMethodRef->pMethod = pJavaClass->FindMethod(strName,strDescriptor);
			if(pMethodRef->pMethod == NULL)
			{
				s = "Can't find method \"";
				s += (char*)strName;
				s += "\" in class \"";
				s += pJavaClass->GetName();
				s += "\".";
			}
			break;

		case 11: //interface method ref			
			pInterfaceMethodRef = (CONSTANT_InterfaceMethodRef*)m_ppConstantPool[i];

			//if method belongs to current class simply update reference
			if(pInterfaceMethodRef->iClassIndex == m_iThisClass)
				pJavaClass = this;
			else
			{
				//or get class full name and try to load it from file
				pClass = (CONSTANT_Class*)m_ppConstantPool[pInterfaceMethodRef->iClassIndex];
				pName = ((CONSTANT_Utf8*)m_ppConstantPool[pClass->iIndex])->pData;

				//could be array def. we want only class name
				pName = ParseObjectName(pName);
				if(pName == NULL)
					break;

				//update ref
				pJavaClass = m_pOwner->LoadClass((const char*)pName);
				if(pJavaClass == NULL)
				{
					s = "Dependent class \"";
					s += (char*)pName;
					s += "\" failed to load.";
					throw s;
				}
			}

			//now we have class pointer
			//get method's name and descriptor
			pNameAndType = (CONSTANT_NameAndType*)m_ppConstantPool[pInterfaceMethodRef->iNameAndTypeIndex];
			strName = ((CONSTANT_Utf8*)m_ppConstantPool[pNameAndType->iNameIndex])->pData;
			strDescriptor = ((CONSTANT_Utf8*)m_ppConstantPool[pNameAndType->iDescriptorIndex])->pData;

			//search class for method and update reference
			pInterfaceMethodRef->pMethod = pJavaClass->FindMethod(strName,strDescriptor);
			if(pInterfaceMethodRef->pMethod == NULL)
			{
				s = "Can't find method \"";
				s += (char*)strName;
				s += "\" in class \"";
				s += pJavaClass->GetName();
				s += "\".";
			}
			break;

		case 8: //string
			//nothing to resolve here
			break;

		case 3: //integer
			//nothing to resolve here
			break;

		case 4: //float
			//nothing to resolve here
			break;

		case 12: //name and type
			//nothing to resolve here
			break;

		case 1: //utf 8
			//nothing to resolve here		
			break;

		case 5: //long 
			++i;//takes two entries in constant pool, so skip next
			break;

		case 6: //double
			++i;//takes two entries in constant pool, so skip next
			break;
		}
	}

	//now compute class size
	u2 iSize = 0;
	u2 iStaticSize = 0;

	if(m_pSuperClass != NULL)
	{
		//get size of super classes
		iSize = (u2)m_pSuperClass->GetSize();
		//iStaticSize = (u2)m_pSuperClass->GetStaticSize();
	}

	//for each field
	for(int i = 0;i < m_iFieldsCount;++i)
	{
		FIELD_INFO *f = &m_pFieldInfos[i];

		u1 ch = *((CONSTANT_Utf8*)m_ppConstantPool[f->iDescriptorIndex])->pData;

		//check descriptor for data type
		switch(ch)
		{
		case 'B': //byte
			f->iSize = sizeof(jByte);
			break;
		case 'C': //char
			f->iSize = sizeof(jChar);
			break;
		case 'D':
			f->iSize = sizeof(jDouble);
			break;
		case 'F':
			f->iSize = sizeof(jFloat);
			break;
		case 'I':
			f->iSize = sizeof(jInt);
			break;
		case 'J':
			f->iSize = sizeof(jLong);
			break;
		case 'L':
			f->iSize = sizeof(void*);
			break;
		case 'S':
			f->iSize = sizeof(jShort);
			break;
		case 'Z':
			f->iSize = sizeof(bool);
			break;
		case '[':
			f->iSize = sizeof(void*);
			break;
		};

		//update size (instance and static)
		if(f->iAccessFlags & ACC_STATIC)
		{
			f->iOffset = iStaticSize;
			iStaticSize += f->iSize;
		}
		else
		{
			f->iOffset = iSize;
			iSize += f->iSize;
		}

		//is field mem reference?
		if(ch == '[' || ch == 'L')
			f->bReference = true;
		else
			f->bReference = false;
	}

	this->m_iSize = iSize;
	this->m_iStaticSize = iStaticSize;
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::InitializeStaticData()
//-----------------------------------------------------------------------------
void cJavaClass::InitializeStaticData()
{
	if(m_iStaticSize == 0)
		return;

	//initialize simple static constants (static finale)
	//complex fields (like arrays) are initialized in class init method (clinit)

	m_pStaticData = new u1[m_iStaticSize];
	//zero mem
	memset(m_pStaticData,0,sizeof(u1) * m_iStaticSize);
	//init constants

	for(int i = 0;i < m_iFieldsCount;++i)
	{
		FIELD_INFO *f = &m_pFieldInfos[i];

		//not a constant
		if(f->iConstantValueIndex == 0)
			continue;
		
		u1 ch = *((CONSTANT_Utf8*)m_ppConstantPool[f->iDescriptorIndex])->pData;
		u1 *ptr;

		short s;
		char c;
		//copy memory based on data type
		switch(ch)
		{
		case 'B': //byte
		case 'Z':
			ptr = m_pStaticData + f->iOffset;
			c = (jByte)((CONSTANT_Integer*)m_ppConstantPool[f->iConstantValueIndex])->iInteger;
			memcpy(ptr,&c,sizeof(jByte));
			break;
		case 'I':
			ptr = m_pStaticData + f->iOffset;
			memcpy(ptr,&((CONSTANT_Integer*)m_ppConstantPool[f->iConstantValueIndex])->iInteger,sizeof(jInt));
			break;
		case 'S':
		case 'C':
			ptr = m_pStaticData + f->iOffset;
			s = (jShort)((CONSTANT_Integer*)m_ppConstantPool[f->iConstantValueIndex])->iInteger;
			memcpy(ptr,&s,sizeof(jShort));
			break;
		
		case 'F':
			ptr = m_pStaticData + f->iOffset;
			memcpy(ptr,&((CONSTANT_Float*)m_ppConstantPool[f->iConstantValueIndex])->fFloat,sizeof(jFloat));
			break;

		case 'J':
			ptr = m_pStaticData + f->iOffset;
			memcpy(ptr,&((CONSTANT_Long*)m_ppConstantPool[f->iConstantValueIndex])->iHighBytes,sizeof(u4));
			ptr += sizeof(u4);
			memcpy(ptr,&((CONSTANT_Long*)m_ppConstantPool[f->iConstantValueIndex])->iLowBytes,sizeof(u4));
			break;

		case 'D':
			ptr = m_pStaticData + f->iOffset;
			memcpy(ptr,&((CONSTANT_Double*)m_ppConstantPool[f->iConstantValueIndex])->iHighBytes,sizeof(u4));
			ptr += sizeof(u4);
			memcpy(ptr,&((CONSTANT_Double*)m_ppConstantPool[f->iConstantValueIndex])->iLowBytes,sizeof(u4));
			break;
		};
	}
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::FindMethod()
//-----------------------------------------------------------------------------
METHOD_INFO* cJavaClass::FindMethod(u1 *strName,u1 *strDescriptor)
{
	//go through all methods
	for(int i = 0;i < this->m_iMethodsCount;++i)
	{
		u1 *strMethodName = ((CONSTANT_Utf8*)m_ppConstantPool[m_pMethodInfos[i].iNameIndex])->pData;
		u1 *strMethodDescriptor = ((CONSTANT_Utf8*)m_ppConstantPool[m_pMethodInfos[i].iDescriptorIndex])->pData;

		//look for same name and descriptor
		if((0 == strcmp((const char*)strName,(const char*)strMethodName)) &&
		   (0 == strcmp((const char*)strDescriptor,(const char*)strMethodDescriptor)))
		{
			return &m_pMethodInfos[i];
		}
	}

	//if nothing found, look in super class
	if(m_pSuperClass != NULL)
		return m_pSuperClass->FindMethod(strName,strDescriptor);
	return NULL;
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::FindField()
//-----------------------------------------------------------------------------
FIELD_INFO* cJavaClass::FindField(u1 *strName,u1 *strDescriptor)
{
	//for all fields
	for(int i = 0;i < this->m_iFieldsCount;++i)
	{
		u1 *strFieldName = ((CONSTANT_Utf8*)m_ppConstantPool[m_pFieldInfos[i].iNameIndex])->pData;
		u1 *strFieldDescriptor = ((CONSTANT_Utf8*)m_ppConstantPool[m_pFieldInfos[i].iDescriptorIndex])->pData;

		//look for same name and descriptor
		if((0 == strcmp((const char*)strName,(const char*)strFieldName)) &&
		   (0 == strcmp((const char*)strDescriptor,(const char*)strFieldDescriptor)))
		{
			return &m_pFieldInfos[i];
		}
	}

	//if nothing found, look in super class
	if(m_pSuperClass != NULL)
		return m_pSuperClass->FindField(strName,strDescriptor);
	return NULL;
}

//-----------------------------------------------------------------------------
// Name: cJavaClass::LoadFromFile()
//-----------------------------------------------------------------------------
void cJavaClass::LoadFromFile(const char *strFilename)
{
	//load class from java class file
	FILE *f;

	f = fopen(strFilename,"rb");
	if(f == 0)
		throw std::string("File not found.");
	
	//check magic value
	m_iMagic = ReadU4(f);
	if(m_iMagic != 0xCAFEBABE)
		throw std::string("Invalid file (wrong magic constant)");

	//read version
	m_iMinor = ReadU2(f);
	m_iMajor = ReadU2(f);

	//read constant pool
	LoadConstantPool(f);

	//DEBUG
	//Debug_PrintConstantPool();

	m_iAccessFlags = ReadU2(f);
	m_iThisClass = ReadU2(f);
	m_iSuperClass = ReadU2(f);

	//read interfaces
	LoadInterfaces(f);

	//read fields
	LoadFieldInfos(f);	

	//read methods
	LoadMethodInfos(f);

	fclose(f);

	//find class name and save it
	CONSTANT_Class *pClass = (CONSTANT_Class*)m_ppConstantPool[this->m_iThisClass];
	u1 *pName = ((CONSTANT_Utf8*)m_ppConstantPool[pClass->iIndex])->pData;

	m_strThisClass = pName;

	//resolve references -> speed up purposes
	//this->ResolveSymbolicReferences();
}