class cJavaClass;
class cJVM;
struct METHOD_INFO;
struct FIELD_INFO;
struct STACK_ITEM;

#include <vector>
#include <map>
#include <string>

//-----------------------------------------------------------------------------
//  Name: struct CONSTANT_Class
//! Represents Class constant in constant_pool
//-----------------------------------------------------------------------------
typedef struct CONSTANT_Class
{
	u1	iTag;
	u2	iIndex;

	//shortcut to class 
	cJavaClass	*pClass;

	CONSTANT_Class(u1 tag,u2 index) : iTag(tag), iIndex(index), pClass(NULL) {}

	void PrintInfo() { printf("Class, Index: %d\n",iIndex); }
} CONSTANT_Class;

//-----------------------------------------------------------------------------
//  Name: struct CONSTANT_FieldRef
//! Represents Field reference constant in constant_pool
//-----------------------------------------------------------------------------
typedef struct CONSTANT_FieldRef
{
	u1	iTag;
	u2	iClassIndex;
	u2	iNameAndTypeIndex;	

	//shortcut to field
	FIELD_INFO	*pField;

	CONSTANT_FieldRef(u1 tag,u2 classIndex,u2 nameAndTypeIndex) 
		: iTag(tag), iClassIndex(classIndex), iNameAndTypeIndex(nameAndTypeIndex), pField(NULL) {}

	void PrintInfo() { printf("Field Ref, ClassIndex: %d, NameAndTypeIndex: %d\n",iClassIndex,iNameAndTypeIndex); }
} CONSTANT_FieldRef;

//-----------------------------------------------------------------------------
//  Name: struct CONSTANT_MethodRef
//! Represents Method reference constant in constant_pool
//-----------------------------------------------------------------------------
typedef struct CONSTANT_MethodRef
{
	u1	iTag;
	u2	iClassIndex;
	u2	iNameAndTypeIndex;

	//shortcut to method
	METHOD_INFO *pMethod;

	CONSTANT_MethodRef(u1 tag,u2 classIndex,u2 nameAndTypeIndex) 
		: iTag(tag), iClassIndex(classIndex), iNameAndTypeIndex(nameAndTypeIndex), pMethod(NULL) {}

	void PrintInfo() { printf("Method Ref, ClassIndex: %d, NameAndTypeIndex: %d\n",iClassIndex,iNameAndTypeIndex); }
} CONSTANT_MethodRef;

//-----------------------------------------------------------------------------
//  Name: struct CONSTANT_InterfaceMethodRef
//! Represents Interface method reference constant in constant_pool
//-----------------------------------------------------------------------------
typedef struct CONSTANT_InterfaceMethodRef
{
	u1	iTag;
	u2	iClassIndex;
	u2	iNameAndTypeIndex;

	//shortcut to method instance
	METHOD_INFO *pMethod;

	CONSTANT_InterfaceMethodRef(u1 tag,u2 classIndex,u2 nameAndTypeIndex) 
		: iTag(tag), iClassIndex(classIndex), iNameAndTypeIndex(nameAndTypeIndex), pMethod(NULL) {}

	void PrintInfo() { printf("Interface Method Ref, ClassIndex: %d, NameAndTypeIndex: %d\n",iClassIndex,iNameAndTypeIndex); }
} CONSTANT_InterfaceMethodRef;

//-----------------------------------------------------------------------------
//  Name: struct CONSTANT_String
//! Represents String constant in constant_pool
//-----------------------------------------------------------------------------
typedef struct CONSTANT_String
{
	u1	iTag;
	u2	iStringIndex;

	CONSTANT_String(u1 tag,u2 stringIndex) : iTag(tag), iStringIndex(stringIndex) {}

	void PrintInfo() { printf("String, Index: %d\n",iStringIndex); }
} CONSTANT_String;

//-----------------------------------------------------------------------------
//  Name: struct CONSTANT_Integer
//! Represents Integer constant in constant_pool
//-----------------------------------------------------------------------------
typedef struct CONSTANT_Integer
{
	u1	iTag;
	u4	iInteger;

	CONSTANT_Integer(u1 tag, u4 data) : iTag(tag), iInteger(data) {}

	void PrintInfo() { printf("Integer, %d\n",iInteger); }
} CONSTANT_Integer;

//-----------------------------------------------------------------------------
//  Name: struct CONSTANT_Float
//! Represents Float constant in constant_pool
//-----------------------------------------------------------------------------
typedef struct CONSTANT_Float
{
	u1	iTag;
	u4	fFloat;

	CONSTANT_Float(u1 tag, u4 data) : iTag(tag), fFloat(data) {}

	void PrintInfo() { printf("Float, %f\n",fFloat); }
} CONSTANT_Float;

//-----------------------------------------------------------------------------
//  Name: struct CONSTANT_Long
//! Represents Long constant in constant_pool
//-----------------------------------------------------------------------------
typedef struct CONSTANT_Long
{
	u1	iTag;
	u4	iHighBytes;
	u4	iLowBytes;

	CONSTANT_Long(u1 tag, u4 highBytes, u4 lowBytes) : iTag(tag), iHighBytes(highBytes), iLowBytes(lowBytes) {}

	void PrintInfo() { printf("Long, %lld\n",(long long)((long long)iHighBytes << 32) | iLowBytes); }
} CONSTANT_Long;

//-----------------------------------------------------------------------------
//  Name: struct CONSTANT_Double
//! Represents Double constant in constant_pool
//-----------------------------------------------------------------------------
typedef struct CONSTANT_Double
{
	u1	iTag;
	u4	iHighBytes;
	u4	iLowBytes;

	CONSTANT_Double(u1 tag, u4 highBytes, u4 lowBytes) : iTag(tag), iHighBytes(highBytes), iLowBytes(lowBytes) {}

	void PrintInfo() 
	{
		long long l = (long long)(((long long)iHighBytes << 32) | iLowBytes);
		double d;
		memcpy(&d,&l,sizeof(double));
		printf("Double, %f\n",d); 
	}
} CONSTANT_Double;

//-----------------------------------------------------------------------------
//  Name: struct CONSTANT_NameAndType
//! Represents Name and Type constant in constant_pool
//-----------------------------------------------------------------------------
typedef struct CONSTANT_NameAndType
{
	u1	iTag;
	u2	iNameIndex;
	u2	iDescriptorIndex;

	CONSTANT_NameAndType(u1 tag,u2 nameIndex, u2 descriptorIndex)
		: iTag(tag), iNameIndex(nameIndex), iDescriptorIndex(descriptorIndex) {}

	void PrintInfo() { printf("Name And Type, NameIndex: %d, DescriptorIndex: %d\n",iNameIndex,iDescriptorIndex); }
} CONSTANT_NameAndType;

//-----------------------------------------------------------------------------
//  Name: struct CONSTANT_Utf8
//! Represents Utf8 constant in constant_pool
//-----------------------------------------------------------------------------
typedef struct CONSTANT_Utf8
{
	u1	iTag;
	u2	iLength;
	u1	*pData;

public:
	CONSTANT_Utf8(u1 tag,u2 length)
		: iTag(tag), iLength(length), pData(NULL) {}
	
	~CONSTANT_Utf8()
	{
		if(pData != NULL)
			delete[] pData;
	}

	void PrintInfo() { printf("Utf8, %s\n",pData); }
} CONSTANT_Utf8;

//-----------------------------------------------------------------------------
//  Name: struct ATTRIBUTE_INFO
//! Holds information about one attribute
//-----------------------------------------------------------------------------
typedef struct ATTRIBUTE_INFO
{
	u2	iNameIndex;
	u4	iLength;
	u1	*pInfo;

	ATTRIBUTE_INFO() { pInfo = NULL; }
	~ATTRIBUTE_INFO() {	SAFE_DELETE_ARRAY(pInfo); }
} ATTRIBUTE_INFO;

//-----------------------------------------------------------------------------
//  Name: struct EXCEPTION_INFO
//! Holds information about exception
//-----------------------------------------------------------------------------
typedef struct EXCEPTION_INFO
{
	u2	iStartPC;
	u2	iEndPC;
	u2	iHandlerPC;
	u2	iCatchType;

	EXCEPTION_INFO()
	{
		iStartPC = 0;
		iEndPC = 0;
		iHandlerPC = 0;
		iCatchType = 0;
	}

} EXCEPTION_INFO;

//-----------------------------------------------------------------------------
//  Name: struct CODE_ATTRIBUTE
//! Holds information about method's byte code
//-----------------------------------------------------------------------------
typedef struct CODE_ATTRIBUTE
{
	u2	iAttributeNameIndex;
	u4	iAttributeLength;
	u2	iMaxStack;
	u2	iMaxLocals;
	u4	iCodeLength;
	u1	*pCode; //bytecode
	u2	iExceptionTableLength;

	EXCEPTION_INFO	*pExceptionTable;

	u2	iAttributesCount;

	ATTRIBUTE_INFO	*pAttributeInfos;

	CODE_ATTRIBUTE()
	{
		pCode = NULL;
		pExceptionTable = NULL;
		pAttributeInfos = NULL;
	}

	~CODE_ATTRIBUTE()
	{
		if(pCode != NULL) delete[] pCode;
		if(pExceptionTable != NULL) delete[] pExceptionTable;
		if(pAttributeInfos != NULL) delete[] pAttributeInfos;
	}

} CODE_ATTRIBUTE;

//-----------------------------------------------------------------------------
//  Name: struct FIELD_INFO
//! Holds information about one field
//-----------------------------------------------------------------------------
typedef struct FIELD_INFO
{
	//owner class
	cJavaClass *pOwner;

	u2	iAccessFlags;
	u2	iNameIndex;
	u2	iDescriptorIndex;
	u2	iAttributesCount;

	ATTRIBUTE_INFO *pAttributeInfos;

	//if field is constant, then this var stores index into constant pool where value of this const is, zero otherwise
	u2	iConstantValueIndex;
	//offset in bytes from beginning of the class
	u2	iOffset; 
	//size in bytes
	u2	iSize; 
	//is field reference or value type
	bool bReference; 

	FIELD_INFO() 
	{ 
		pAttributeInfos = NULL;
		iConstantValueIndex = 0;
		pOwner = NULL;
		bReference = false;
		iOffset = 0;
		iSize = 0;
	}
	~FIELD_INFO() { if(pAttributeInfos != NULL) delete[] pAttributeInfos; }
} FIELD_INFO;

//-----------------------------------------------------------------------------
//  Name: struct METHOD_INFO
//! Holds information about one method
//-----------------------------------------------------------------------------
typedef struct METHOD_INFO
{
	//owner
	cJavaClass	*pOwner;

	u2	iAccessFlags;
	u2	iNameIndex;
	u2	iDescriptorIndex;
	u2	iAttrbutesCount;
	ATTRIBUTE_INFO	*pAttributeInfos;

	//max stack size for this method
	u2	iMaxStack;
	//number of local variables
	u2	iMaxLocals;
	//bytecode length
	u4	iCodeLength;
	//bytecode
	u1	*pCode;
	//exception table length
	u2	iExceptionTableLength;

	//Exception infos
	EXCEPTION_INFO	*pExceptionTable;

	//bytecode attribute count
	u2	iCodeAttributesCount;
	//bytecode attributes
	ATTRIBUTE_INFO	*pCodeAttributeInfos;

	//number of stack values arguments take
	u2	iNumArguments;

	METHOD_INFO()
	{
		pOwner = NULL;
		pAttributeInfos = NULL;
		pCode = NULL;
		pExceptionTable = NULL;
		pCodeAttributeInfos = NULL;
		iNumArguments = 0;
		iAccessFlags = 0;
		iNameIndex = 0;
		iDescriptorIndex = 0;
		iAttrbutesCount = 0;
		iMaxStack = 0;
		iMaxLocals = 0;
		iCodeLength = 0;
		iExceptionTableLength = 0;
		iCodeAttributesCount = 0;
	}

	~METHOD_INFO()
	{
		if(pAttributeInfos != NULL) delete[] pAttributeInfos;
		if(pCode != NULL) delete[] pCode;
		if(pExceptionTable != NULL) delete[] pExceptionTable;
		if(pAttributeInfos != NULL) delete[] pCodeAttributeInfos;
	}
} METHOD_INFO;

//-----------------------------------------------------------------------------
//  Name: class cJavaClass
//! Represents java class
//-----------------------------------------------------------------------------
class cJavaClass
{
	friend cJVM;
private:
	u4	m_iMagic;
	u2	m_iMinor;
	u2	m_iMajor;

	//num entries in constant pool
	u2	m_iConstantPoolCount;
	//constant pool
	void	**m_ppConstantPool;

	u2	m_iAccessFlags;
	u2	m_iThisClass;
	u2	m_iSuperClass;

	//number of interfaces
	u2	m_iInterfaceCount;
	//interfaces
	u2	*m_pInterfaces;	

	//number of fields
	u2	m_iFieldsCount;
	//fields
	FIELD_INFO	*m_pFieldInfos;
	
	//number of methods
	u2	m_iMethodsCount;
	//method infos
	METHOD_INFO	*m_pMethodInfos;

	//pointer to main method if present
	METHOD_INFO	*m_pMainMethod;
	//pointer to class initialization method if present
	METHOD_INFO	*m_pClassInit;

	//class members size in bytes (including super classes(parents))
	u2	m_iSize;
	//class static members size in bytes (including super classes(parents))
	u2	m_iStaticSize;

	//pointer to super class or null for Object
	cJavaClass	*m_pSuperClass;

	//class name (char*)
	u1* m_strThisClass;

	//pointer to parent virtual machine
	cJVM *m_pOwner;

	//classe's static data
	u1	*m_pStaticData;

	//Is class initialized - currently unused
	bool m_bInitialized;
	
private:
	//Loads constant pool
	void LoadConstantPool(FILE *f);
	//Loads method infos
	void LoadMethodInfos(FILE *f);
	//Loads interfaces
	void LoadInterfaces(FILE *f);
	//Loads field infos
	void LoadFieldInfos(FILE *f);
	//Get size of arguments in stack values
	u2 ParseAgrumentCount(u1 *strDescriptor);
	//Initializes static data
	void InitializeStaticData();
	//Replace references to constant pool with direct pointers for speed up purposes
	void ResolveSymbolicReferences();
	//Load from jaca class file
	void LoadFromFile(const char *strFilename);
	//Parse object name
	u1* ParseObjectName(u1 *name);
	//Parse array definition
	cJavaClass* ParseArrayDefinition(u2 iConstantIndex,u1 &iDimensions,u1 &iArrayItemSize);

public:
	cJavaClass(cJVM *pOwner);
	~cJavaClass();

	const char* GetName() { return (char*)m_strThisClass; }
	//Debug const pool output (stdout)
	void Debug_PrintConstantPool();
	//Gets size of class
	u2 GetSize() { return m_iSize; }
	//Gets size of static members
	u2 GetStaticSize() { return m_iStaticSize; }
	//Gets main method
	METHOD_INFO* GetMainMethod() { return m_pMainMethod; }
	//Find specific method
	METHOD_INFO* FindMethod(u1 *strName,u1 *strDescriptor);
	//Find specific field
	FIELD_INFO* FindField(u1 *strName,u1 *strDescriptor);
	//Get class from pool reference
	cJavaClass* GetClassFromPool(u2 iIndex);
	//Get method from pool reference
	METHOD_INFO* GetMethodFromPool(u2 iIndex);
	//Get interface method from pool reference
	METHOD_INFO* GetInterfaceMethodFromPool(u2 iIndex);
	//Get Field from pool reference
	FIELD_INFO* GetFieldFromPool(u2 iIndex);
	//Get interface method name and desc
	void GetInterfaceMethodNameAndDesc(u2 iIndex,u1 **ppName, u1 **ppDesc);
	//Set static value
	void SetStaticValue(u2 iOffset, void *pData, u2 iSize);
	//Get static value
	void GetStaticValue(u2 iOffset, void *pData, u2 iSize);
	//Get 8-byte value from constant pool
	bool GetDoubleOrLongFromConstantPool(u2 iIndex,u4 &iHighBytes, u4 &iLowBytes);
	//Get 4-byte value from constant pool
	bool GetFloatOrIntFromConstantPool(u2 iIndex,u4 *pData);
	//Get number of fields
	u2 GetNumFields() { return m_iFieldsCount; }
	//Get field by specified index.
	FIELD_INFO* GetFieldInfo(int iIndex) { if(iIndex >= m_iFieldsCount) return NULL; return &m_pFieldInfos[iIndex]; }
	//get super class
	cJavaClass* GetSuperClass() { return m_pSuperClass; }
	//is interface
	bool IsInterface()
	{
		if(m_iAccessFlags & ACC_INTERFACE)
			return true;
		return false;
	}
	//get num interfaces
	u4 GetNumInterfaces() { return m_iInterfaceCount; }
	//get interface
	cJavaClass* GetInterface(u4 iIndex)
	{
		return ((CONSTANT_Class*)m_ppConstantPool[m_pInterfaces[iIndex]])->pClass;
	}
};