#define _CRT_SECURE_NO_WARNINGS

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <map>
#include <string>

#include "Common.h"
#include "OpCodes.h"
#include "Class.h"
#include "Heap.h"
#include "Stack.h"
#include "JVM.h"

//----------------------------------------------------------------------------
// TESTING ONLY
//----------------------------------------------------------------------------
int main()
{
	//create jvm instance
	cJVM	jvm;
	//initialize (stack size, heap size)
	jvm.Initialize(128,700);
	//load class with main method
	jvm.LoadMain("myClass.class");

	//run jvm
	jvm.Run();
}

//-----------------------------------------------------------------------------
// Name: cJVM::cJVM()
//-----------------------------------------------------------------------------
cJVM::cJVM()
{
	m_pPC = NULL;
	m_bError = false;
	m_Heap.SetOwner(this);

	for(int i = 0;i < INST_NUMINST;++i)
		Instructions[i] = NULL;

	//assign instruction functions
	Instructions[INST_NEW] = &cJVM::Inst_New;
	Instructions[INST_DUP] = &cJVM::Inst_Dup;
	Instructions[INST_INVOKESPECIAL] = &cJVM::Inst_InvokeSpecial;
	Instructions[INST_INVOKEVIRTUAL] = &cJVM::Inst_InvokeVirtual;
	Instructions[INST_INVOKEINTERFACE] = &cJVM::Inst_InvokeInterface;
	Instructions[INST_DLOAD] = &cJVM::Inst_DLoad;
	Instructions[INST_DLOAD_1] = &cJVM::Inst_DLoad_1;
	Instructions[INST_ALOAD] = &cJVM::Inst_ALoad;
	Instructions[INST_ALOAD_0] = &cJVM::Inst_ALoad_0;
	Instructions[INST_ALOAD_1] = &cJVM::Inst_ALoad_1;
	Instructions[INST_ALOAD_2] = &cJVM::Inst_ALoad_2;
	Instructions[INST_ALOAD_3] = &cJVM::Inst_ALoad_3;
	Instructions[INST_RETURN] = &cJVM::Inst_Return;
	Instructions[INST_IRETURN] = &cJVM::Inst_IReturn;
	Instructions[INST_ARETURN] = &cJVM::Inst_AReturn;
	Instructions[INST_ASTORE] = &cJVM::Inst_AStore;
	Instructions[INST_ASTORE_0] = &cJVM::Inst_AStore_0;
	Instructions[INST_ASTORE_1] = &cJVM::Inst_AStore_1;
	Instructions[INST_ASTORE_2] = &cJVM::Inst_AStore_2;
	Instructions[INST_ASTORE_3] = &cJVM::Inst_AStore_3;
	Instructions[INST_ISTORE] = &cJVM::Inst_IStore;
	Instructions[INST_ISTORE_1] = &cJVM::Inst_IStore_1;
	Instructions[INST_ISTORE_0] = &cJVM::Inst_IStore_0;
	Instructions[INST_ISTORE_2] = &cJVM::Inst_IStore_2;
	Instructions[INST_ISTORE_3] = &cJVM::Inst_IStore_3;
	Instructions[INST_ICONST_0] = &cJVM::Inst_IConst_0;
	Instructions[INST_ICONST_1] = &cJVM::Inst_IConst_1;
	Instructions[INST_ICONST_2] = &cJVM::Inst_IConst_2;
	Instructions[INST_ICONST_3] = &cJVM::Inst_IConst_3;
	Instructions[INST_ICONST_4] = &cJVM::Inst_IConst_4;
	Instructions[INST_ICONST_5] = &cJVM::Inst_IConst_5;
	Instructions[INST_ILOAD] = &cJVM::Inst_ILoad;
	Instructions[INST_ILOAD_0] = &cJVM::Inst_ILoad_0;
	Instructions[INST_ILOAD_1] = &cJVM::Inst_ILoad_1;
	Instructions[INST_ILOAD_2] = &cJVM::Inst_ILoad_2;
	Instructions[INST_ILOAD_3] = &cJVM::Inst_ILoad_3;
	Instructions[INST_IADD] = &cJVM::Inst_IAdd;
	Instructions[INST_POP] = &cJVM::Inst_Pop;
	Instructions[INST_IASTORE] = &cJVM::Inst_IAStore;
	Instructions[INST_BIPUSH] = &cJVM::Inst_BIPush;
	Instructions[INST_PUTSTATIC] = &cJVM::Inst_PutStatic;
	Instructions[INST_GETSTATIC] = &cJVM::Inst_GetStatic;
	Instructions[INST_PUTFIELD] = &cJVM::Inst_PutField;
	Instructions[INST_GETFIELD] = &cJVM::Inst_GetField;
	Instructions[INST_NEWARRAY] = &cJVM::Inst_NewArray;
	Instructions[INST_GOTO] = &cJVM::Inst_GoTo;
	Instructions[INST_BALOAD] = &cJVM::Inst_BALoad;
	Instructions[INST_IINC] = &cJVM::Inst_IInc;
	Instructions[INST_ARRAYLENGTH] = &cJVM::Inst_ArrayLength;
	Instructions[INST_IF_ICMPLT] = &cJVM::Inst_If_ICMPLT;
	Instructions[INST_SIPUSH] = &cJVM::Inst_SIPush;
	Instructions[INST_BASTORE] = &cJVM::Inst_BAStore;
	Instructions[INST_IALOAD] = &cJVM::Inst_IALoad;
	Instructions[INST_AALOAD] = &cJVM::Inst_AALoad;
	Instructions[INST_LDC2_W] = &cJVM::Inst_LDC2_W;
	Instructions[INST_DADD] = &cJVM::Inst_DAdd;
	Instructions[INST_DRETURN] = &cJVM::Inst_DReturn;
	Instructions[INST_D2I] = &cJVM::Inst_D2I;
	Instructions[INST_ANEWARRAY] = &cJVM::Inst_ANewArray;
	Instructions[INST_DCONST_0] = &cJVM::Inst_DConst_0;
	Instructions[INST_DCONST_1] = &cJVM::Inst_DConst_1;	
	Instructions[INST_AASTORE] = &cJVM::Inst_AAStore;	
	Instructions[INST_INVOKESTATIC] = &cJVM::Inst_InvokeStatic;
	Instructions[INST_DSTORE_1] = &cJVM::Inst_DStore_1;
	Instructions[INST_ACONST_NULL] = &cJVM::Inst_AConst_Null;

	Instructions[INST_DLOAD_0] = &cJVM::Inst_DLoad_0;
	Instructions[INST_DLOAD_1] = &cJVM::Inst_DLoad_1;
	Instructions[INST_DLOAD_2] = &cJVM::Inst_DLoad_2;
	Instructions[INST_DLOAD_3] = &cJVM::Inst_DLoad_3;
	Instructions[INST_FRETURN] = &cJVM::Inst_FReturn;
	Instructions[INST_DRETURN] = &cJVM::Inst_DReturn;
	Instructions[INST_DSTORE] = &cJVM::Inst_DStore;
	Instructions[INST_FSTORE] = &cJVM::Inst_FStore;
	Instructions[INST_FSTORE_0] = &cJVM::Inst_FStore_0;
	Instructions[INST_FSTORE_1] = &cJVM::Inst_FStore_1;
	Instructions[INST_FSTORE_2] = &cJVM::Inst_FStore_2;
	Instructions[INST_FSTORE_3] = &cJVM::Inst_FStore_3;
	Instructions[INST_DSTORE_0] = &cJVM::Inst_DStore_0;
	Instructions[INST_DSTORE_1] = &cJVM::Inst_DStore_1;
	Instructions[INST_DSTORE_2] = &cJVM::Inst_DStore_2;
	Instructions[INST_DSTORE_3] = &cJVM::Inst_DStore_3;
	Instructions[INST_FLOAD] = &cJVM::Inst_FLoad;
	Instructions[INST_FLOAD_0] = &cJVM::Inst_FLoad_0;
	Instructions[INST_FLOAD_1] = &cJVM::Inst_FLoad_1;
	Instructions[INST_FLOAD_2] = &cJVM::Inst_FLoad_2;
	Instructions[INST_FLOAD_3] = &cJVM::Inst_FLoad_3;
	Instructions[INST_FADD] = &cJVM::Inst_FAdd;
	Instructions[INST_FSUB] = &cJVM::Inst_FSub;
	Instructions[INST_FDIV] = &cJVM::Inst_FDiv;
	Instructions[INST_FMUL] = &cJVM::Inst_FMul;
	Instructions[INST_FREM] = &cJVM::Inst_FRem;
	Instructions[INST_FNEG] = &cJVM::Inst_FNeg;
	Instructions[INST_FASTORE] = &cJVM::Inst_FAStore;
	Instructions[INST_DASTORE] = &cJVM::Inst_DAStore;
	Instructions[INST_GOTO_W] = &cJVM::Inst_GoTo_W;
	Instructions[INST_CALOAD] = &cJVM::Inst_CALoad;
	Instructions[INST_BALOAD] = &cJVM::Inst_BALoad;
	Instructions[INST_FALOAD] = &cJVM::Inst_FALoad;
	Instructions[INST_DALOAD] = &cJVM::Inst_DALoad;
	Instructions[INST_CASTORE] = &cJVM::Inst_CAStore;
	Instructions[INST_DSUB] = &cJVM::Inst_DSub;
	Instructions[INST_DREM] = &cJVM::Inst_DRem;
	Instructions[INST_DNEG] = &cJVM::Inst_DNeg;
	Instructions[INST_DMUL] = &cJVM::Inst_DMul;
	Instructions[INST_DDIV] = &cJVM::Inst_DDiv;
	Instructions[INST_D2F] = &cJVM::Inst_D2F;
	Instructions[INST_D2L] = &cJVM::Inst_D2L;
	Instructions[INST_DCMPL] = &cJVM::Inst_DCMPL;
	Instructions[INST_DCMPG] = &cJVM::Inst_DCMPG;
	Instructions[INST_FCMPL] = &cJVM::Inst_FCMPL;
	Instructions[INST_FCMPG] = &cJVM::Inst_FCMPG;
	Instructions[INST_F2D] = &cJVM::Inst_F2D;
	Instructions[INST_F2L] = &cJVM::Inst_F2L;
	Instructions[INST_F2I] = &cJVM::Inst_F2I;
	Instructions[INST_FCONST_0] = &cJVM::Inst_FConst_0;
	Instructions[INST_FCONST_1] = &cJVM::Inst_FConst_1;
	Instructions[INST_FCONST_2] = &cJVM::Inst_FConst_2;
	Instructions[INST_I2B] = &cJVM::Inst_I2B;
	Instructions[INST_I2C] = &cJVM::Inst_I2C;
	Instructions[INST_I2D] = &cJVM::Inst_I2D;
	Instructions[INST_I2F] = &cJVM::Inst_I2F;
	Instructions[INST_I2L] = &cJVM::Inst_I2L;
	Instructions[INST_I2S] = &cJVM::Inst_I2S;
	Instructions[INST_ICONST_M1] = &cJVM::Inst_IConst_M1;
	Instructions[INST_IAND] = &cJVM::Inst_IAnd;
	Instructions[INST_ISHL] = &cJVM::Inst_ISHL;
	Instructions[INST_ISHR] = &cJVM::Inst_ISHR;
	Instructions[INST_IUSHR] = &cJVM::Inst_IUSHR;
	Instructions[INST_IREM] = &cJVM::Inst_IRem;
	Instructions[INST_IOR] = &cJVM::Inst_IOr;
	Instructions[INST_IXOR] = &cJVM::Inst_IXOr;
	Instructions[INST_INEG] = &cJVM::Inst_INeg;
	Instructions[INST_IMUL] = &cJVM::Inst_IMul;
	Instructions[INST_ISUB] = &cJVM::Inst_ISub;
	Instructions[INST_IDIV] = &cJVM::Inst_IDiv;
	Instructions[INST_IF_ICMPEQ] = &cJVM::Inst_If_ICMPEQ;
	Instructions[INST_IF_ICMPNE] = &cJVM::Inst_If_ICMPNE;
	Instructions[INST_IF_ICMPLT] = &cJVM::Inst_If_ICMPLT;
	Instructions[INST_IF_ICMPLE] = &cJVM::Inst_If_ICMPLE;
	Instructions[INST_IF_ICMPGT] = &cJVM::Inst_If_ICMPGT;
	Instructions[INST_IF_ICMPGE] = &cJVM::Inst_If_ICMPGE;
	Instructions[INST_IF_ACMPEQ] = &cJVM::Inst_If_ACMPEQ;
	Instructions[INST_IF_ACMPNE] = &cJVM::Inst_If_ACMPNE;
	Instructions[INST_IFEQ] = &cJVM::Inst_IFEQ;
	Instructions[INST_IFNE] = &cJVM::Inst_IFNE;
	Instructions[INST_IFLT] = &cJVM::Inst_IFLT;
	Instructions[INST_IFLE] = &cJVM::Inst_IFLE;
	Instructions[INST_IFGT] = &cJVM::Inst_IFGT;
	Instructions[INST_IFGE] = &cJVM::Inst_IFGE;
	Instructions[INST_IFNONNULL] = &cJVM::Inst_IfNonNull;
	Instructions[INST_IFNULL] = &cJVM::Inst_IfNull;

	Instructions[INST_JSR] = &cJVM::Inst_JSR;
	Instructions[INST_JSR_W] = &cJVM::Inst_JSR_W;
	Instructions[INST_SWAP] = &cJVM::Inst_Swap;
	Instructions[INST_NOP] = &cJVM::Inst_Nop;
	Instructions[INST_LCONST_0] = &cJVM::Inst_LConst_0;
	Instructions[INST_LCONST_1] = &cJVM::Inst_LConst_1;
	Instructions[INST_L2D] = &cJVM::Inst_L2D;
	Instructions[INST_L2F] = &cJVM::Inst_L2F;
	Instructions[INST_L2I] = &cJVM::Inst_L2I;
	Instructions[INST_LALOAD] = &cJVM::Inst_LALoad;
	Instructions[INST_LADD] = &cJVM::Inst_LAdd;
	Instructions[INST_LDIV] = &cJVM::Inst_LDiv;
	Instructions[INST_LSUB] = &cJVM::Inst_LSub;
	Instructions[INST_LSHL] = &cJVM::Inst_LShl;
	Instructions[INST_LSHR] = &cJVM::Inst_LShr;
	Instructions[INST_LUSHR] = &cJVM::Inst_LUShr;
	Instructions[INST_LOR] = &cJVM::Inst_LOr;
	Instructions[INST_LXOR] = &cJVM::Inst_LXOr;
	Instructions[INST_LREM] = &cJVM::Inst_LRem;
	Instructions[INST_LNEG] = &cJVM::Inst_LNeg;
	Instructions[INST_LMUL] = &cJVM::Inst_LMul;
	Instructions[INST_LAND] = &cJVM::Inst_LAnd;
	Instructions[INST_LCMP] = &cJVM::Inst_LCmp;
	Instructions[INST_LDC] = &cJVM::Inst_LDC;
	Instructions[INST_LDC_W] = &cJVM::Inst_LDC_W;
	Instructions[INST_SALOAD] = &cJVM::Inst_SALoad;
	Instructions[INST_LASTORE] = &cJVM::Inst_LAStore;
	Instructions[INST_SASTORE] = &cJVM::Inst_SAStore;
	Instructions[INST_POP2] = &cJVM::Inst_Pop2;
	Instructions[INST_LSTORE_0] = &cJVM::Inst_LStore_0;
	Instructions[INST_LSTORE_1] = &cJVM::Inst_LStore_1;
	Instructions[INST_LSTORE_2] = &cJVM::Inst_LStore_2;
	Instructions[INST_LSTORE_3] = &cJVM::Inst_LStore_3;
	Instructions[INST_LSTORE] = &cJVM::Inst_LStore;
	Instructions[INST_LRETURN] = &cJVM::Inst_LReturn;
	Instructions[INST_LLOAD] = &cJVM::Inst_LLoad;
	Instructions[INST_LLOAD_0] = &cJVM::Inst_LLoad_0;
	Instructions[INST_LLOAD_1] = &cJVM::Inst_LLoad_1;
	Instructions[INST_LLOAD_2] = &cJVM::Inst_LLoad_2;
	Instructions[INST_LLOAD_3] = &cJVM::Inst_LLoad_3;
	Instructions[INST_RET] = &cJVM::Inst_Ret;
	Instructions[INST_MULTIANEWARRAY] = &cJVM::Inst_MultiANewArray;

	Instructions[INST_DUP_X1] = &cJVM::Inst_Dup_X1;
	Instructions[INST_DUP_X2] = &cJVM::Inst_Dup_X2;
	Instructions[INST_DUP2] = &cJVM::Inst_Dup2;
	Instructions[INST_DUP2_X1] = &cJVM::Inst_Dup2_X1;
	Instructions[INST_DUP2_X2] = &cJVM::Inst_Dup2_X2;

	Instructions[INST_CHECKCAST] = &cJVM::Inst_CheckCast;
	Instructions[INST_INSTANCEOF] = &cJVM::Inst_InstanceOf;

	Instructions[INST_WIDE] = &cJVM::Inst_Wide;
}

//-----------------------------------------------------------------------------
// Name: cJVM::~cJVM()
//-----------------------------------------------------------------------------
cJVM::~cJVM()
{
	//make sure to release allocated memory
	FreeAllClasses();
}

//-----------------------------------------------------------------------------
// Name: cJVM::Error()
//-----------------------------------------------------------------------------
void cJVM::Error(char *strError)
{
	//stop execution
	m_pPC = NULL;
	//print error
	printf("%s\n",strError);
	//clear stack
	m_Stack.Clear();
	//raise flag
	m_bError = true;
}

//-----------------------------------------------------------------------------
// Name: cJVM::FreeAllClasses()
//-----------------------------------------------------------------------------
void cJVM::FreeAllClasses()
{
	//get rid of them, one by one...
	for(ClassMapItor itor = m_Classes.begin();itor != m_Classes.end();++itor)
	{
		SAFE_DELETE(itor->second);
	}
	m_Classes.clear();
}

//-----------------------------------------------------------------------------
// Name: cJVM::Initialize()
//-----------------------------------------------------------------------------
bool cJVM::Initialize(u4 iStackSize, u4 iHeapSize)
{
	//one time initialization
	//stack size - in stack values
	//heap size - in bytes
	m_Stack.Resize(iStackSize);
	m_Heap.Init(iHeapSize);
	return true;
}

//-----------------------------------------------------------------------------
// Name: cJVM::LoadMain()
//-----------------------------------------------------------------------------
bool cJVM::LoadMain(const char *strFilename)
{
	//create class instance
	cJavaClass *pClass = new cJavaClass(this);

	try
	{
		//try to load class from file
		pClass->LoadFromFile(strFilename);

		//add class to map
		std::pair<ClassMapItor,bool> pr = m_Classes.insert(ClassMapValue(std::string(strFilename),pClass));
		if(pr.second == false) //unable to insert
		{
			throw std::string("Internal Error");
		}

		//try to resolve constant pool references
		pClass->ResolveSymbolicReferences();
		//init static(constant) data 
		pClass->InitializeStaticData();

		//run classe's initialization method (clinit)
		if(!this->InitializeClass(pClass))
		{
			throw std::string("Class's initialization method failed.");
		}
	}
	catch(std::string strError)
	{
		m_Classes.erase(std::string(strFilename));
		SAFE_DELETE(pClass);
		printf("Unable to load class \"%s\" - %s\n",strFilename,strError.c_str());
		return false;
	}

	//check for main
	if(pClass->m_pMainMethod == NULL)
		return false;

	//prepare main method for running
	//method's frame
	m_Stack.PushMethodFrame(pClass->m_pMainMethod, NULL, 0);
	//program counter
	m_pPC = pClass->m_pMainMethod->pCode;
	//and currnet class
	m_pCurClass = pClass;

	return true;
}

//-----------------------------------------------------------------------------
// Name: cJVM::LoadClass()
//-----------------------------------------------------------------------------
cJavaClass* cJVM::LoadClass(const char *strFilename)
{
	//load class from java class file
	//class names in constant pool don't have extension, so add one
	std::string s = strFilename;
	s += ".class";

	//check if this class is already loaded
	ClassMapItor i = m_Classes.find(s);
	if(i != m_Classes.end())
	{
		return i->second;
	}

	//load class
	cJavaClass *pClass = new cJavaClass(this);

	try
	{
		pClass->LoadFromFile(s.c_str());

		//add class to map
		std::pair<ClassMapItor,bool> pr = m_Classes.insert(ClassMapValue(s,pClass));
		if(pr.second == false) //unable to insert
		{
			//this really shouldn't happen, so just for sure
			throw std::string("Internal Error");
		}

		//try to resolve constant pool references
		pClass->ResolveSymbolicReferences();
		//init static(constant) data 
		pClass->InitializeStaticData();

		//run classe's initialization method (clinit)
		if(!this->InitializeClass(pClass))
		{
			throw std::string("Class's initialization method failed.");
		}
	}
	catch(std::string strError)
	{
		m_Classes.erase(s);
		SAFE_DELETE(pClass);
		printf("Unable to load class \"%s\" - %s\n",strFilename,strError.c_str());
		return NULL;
	}	

	return pClass;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Run()
//-----------------------------------------------------------------------------
bool cJVM::Run()
{
	//start execution prepared metohd
	//first, check if any method was prepared
	if(m_pPC == NULL)
		return false;

	//now, go trought byto code, one instruction at a time
	u1 iOpCode;
	while(m_pPC != NULL)
	{
		//fetch opcode
		iOpCode = *m_pPC;
		//printf("%d\n",iOpCode); //DEBUG
		m_pPC++;

		//this section could be erased after all instrucionts are implemented
		if(this->Instructions[iOpCode] == NULL)
		{ //opcode isn't implemented
			printf("Unknown opcode: %d\n",iOpCode);
			return false;
		}

		//call appropriate function based on opcode
		(this->*Instructions[iOpCode])();
	}

	if(m_bError)
	{
		m_bError = false;
		return false;
	}

	//DEBUG OUTPUT - get rid of it when no longer needed
	printf("\nAnd the result is: %d\n",m_Stack.GetOperand(4).iInt);

	return true;
}


//-----------------------------------------------------------------------------
// Name: cJVM::InitializeClass()
//-----------------------------------------------------------------------------
bool cJVM::InitializeClass(cJavaClass *pClass)
{
	//call classe's initialization method (clinit)
	if(pClass->m_pClassInit == NULL)
		return true;

	//setup stak frame
	m_Stack.PushMethodFrame(pClass->m_pClassInit, m_pPC, 0);
	m_pPC = pClass->m_pClassInit->pCode;
	m_pCurClass = pClass;

	//execute bytecode
	u1 iOpCode;
	while(m_pPC != NULL)
	{
		iOpCode = *m_pPC;
		//printf("%d\n",iOpCode);
		m_pPC++;

		if(this->Instructions[iOpCode] == NULL)
		{ //opcode isn't implemented
			printf("Unknown opcode: %d\n",iOpCode);
			return false;
		}
		(this->*Instructions[iOpCode])();
	}

	if(m_bError)
	{
		m_bError = false;
		return false;
	}

	return true;
}

//-----------------------------------------------------------------------------
// Name: cJVM::ReadU2()
//-----------------------------------------------------------------------------
u2 cJVM::ReadU2() 
{
	//read two byte value from pc
	u1 i1 = *m_pPC;
	++m_pPC;
	u1 i2 = *m_pPC;
	++m_pPC;

	//it's big endian and we need little

	return (u2)((i1 << 8) | i2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::ReadU4()
//-----------------------------------------------------------------------------
u4 cJVM::ReadU4() 
{
	//read two byte value from pc
	u1 i1 = *m_pPC;
	++m_pPC;
	u1 i2 = *m_pPC;
	++m_pPC;
	u1 i3 = *m_pPC;
	++m_pPC;
	u1 i4 = *m_pPC;
	++m_pPC;

	//it's big endian and we need little
	return (u4)((i1 << 24) | (i2 << 16) | (i3 << 8) | i4);
}

//-----------------------------------------------------------------------------
// Name: cJVM::GetDouble()
//-----------------------------------------------------------------------------
void cJVM::GetDouble(jDouble &d,u4 iHighBytes,u4 iLowBytes)
{
	//compute double from high and low bytes
	//just for clarification, long long in C++ is the same as long in java (8-byte)
	jLong l = (jLong)(((jLong)iHighBytes << 32) | iLowBytes);
	memcpy(&d,&l,sizeof(jDouble));
}

//-----------------------------------------------------------------------------
// Name: cJVM::GetLong()
//-----------------------------------------------------------------------------
void cJVM::GetLong(jLong &l,u4 iHighBytes,u4 iLowBytes)
{
	//compute long from high and low bytes
	//just for clarification, long long in C++ is the same as long in java (8-byte)
	l = (jLong)(((jLong)iHighBytes << 32) | iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::GetBytesFromDouble()
//-----------------------------------------------------------------------------
void cJVM::GetBytesFromDouble(jDouble &d,u4 &iHighBytes,u4 &iLowBytes)
{
	//split double to high and low bytes
	u1 *pData = (u1*)&d;

	memcpy(&iLowBytes,pData,sizeof(u4));
	pData += sizeof(u4);
	memcpy(&iHighBytes,pData,sizeof(u4));		
}

//-----------------------------------------------------------------------------
// Name: cJVM::GetBytesFromLong()
//-----------------------------------------------------------------------------
void cJVM::GetBytesFromLong(jLong &l,u4 &iHighBytes,u4 &iLowBytes)
{
	//split long to high and low bytes
	u1 *pData = (u1*)&l;

	memcpy(&iLowBytes,pData,sizeof(u4));
	pData += sizeof(u4);
	memcpy(&iHighBytes,pData,sizeof(u4));		
}

//-----------------------------------------------------------------------------
// Name: cJVM::CreateMultiDimArray()
//-----------------------------------------------------------------------------
OBJECT* cJVM::CreateMultiDimArray(u1 iCurDim, u1 iNumDim, u1 iItemSize, cJavaClass *pItem)
{
	jInt iDimSize = m_Stack.GetOperand(m_Stack.GetOperandStackTop() - iNumDim + iCurDim).iInt;

	if(iCurDim == iNumDim - 1)
	{
		//last dimension. create items		
		return m_Heap.AllocateArray(pItem,iDimSize,iItemSize);
	}
	else
	{
		//
		OBJECT *pObj = m_Heap.AllocateArray(pItem,iDimSize,iItemSize);

		//skip header
		OBJECT **ppObjArray = (OBJECT**)(pObj + 1);

		for(jInt i = 0;i < iDimSize;++i)
		{
			//alocate next dimeensions
			ppObjArray[i] = CreateMultiDimArray(iCurDim + 1,iNumDim,iItemSize,pItem);
			if(ppObjArray[i] == NULL)
				return NULL;
		}

		return pObj;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::CheckCast()
//-----------------------------------------------------------------------------
bool cJVM::CheckCast(OBJECT *pObj, u2 iConstantIndex)
{
	u1 iDimensions,iItemSize;
	cJavaClass *pClass = m_pCurClass->ParseArrayDefinition(iConstantIndex,iDimensions,iItemSize);

	return CheckCastRecursive(pObj,pClass,iDimensions,iItemSize);
}

//-----------------------------------------------------------------------------
// Name: cJVM::CheckCast()
//-----------------------------------------------------------------------------
bool cJVM::CheckCastRecursive(OBJECT *pObj, cJavaClass *pClass, u1 iDimensions, u1 iItemSize)
{
	//for more info: http://java.sun.com/docs/books/jvms/second_edition/html/Instructions2.doc2.html
	if(pObj->iArraySize == 1) //non array
	{
		if(pObj->pClass->IsInterface())
		{
			if(pClass == NULL) return false;
			if(pClass->IsInterface())
			{
				cJavaClass *pC = pObj->pClass;
				while(pC != NULL)
				{
					if(pC == pClass)
						return true;
					pC = pC->GetSuperClass();
				}

				return false;
			}
			else
			{
				if(pClass->GetSuperClass() == NULL)
				{
					//pClass is Object
					return true;
				}
				return false;
			}
		}
		else
		{
			if(pClass == NULL) return false;
			if(pClass->IsInterface())
			{
				for(u4 i = 0;i < pObj->pClass->GetNumInterfaces();++i)
				{
					if(pClass == pObj->pClass->GetInterface(i))
						return true;
				}
				return false;
			}
			else
			{
				if(pClass == NULL) return false;
				cJavaClass *pC = pObj->pClass;
				while(pC != NULL)
				{
					if(pC == pClass)
						return true;
					pC = pC->GetSuperClass();
				}

				return false;
			}
		}
	}
	else
	{
		if(iDimensions == 0) //pClass is non array
		{
			if(pClass == NULL) return false;
			if(pClass->GetSuperClass() == NULL)
			{
				//pClass is Object
				return true;
			}
			return false;
		}
		else
		{
			if(pClass == NULL && pObj->pClass == NULL) //primitive type
			{
				jInt iObjArrayItemSize = (pObj->iSize - sizeof(OBJECT)) / pObj->iArraySize;
				if(iObjArrayItemSize == iItemSize)
					return true;
				return false;
			}
			else if(pClass != NULL && pObj->pClass != NULL)
			{
				return CheckCastRecursive(pObj,pClass,0,0);
			}
			else
			{
				return false;
			}
		}
	}
}


//-----------------------------------------------------------------------------
// Name: cJVM::Inst_New()
//-----------------------------------------------------------------------------
void cJVM::Inst_New()
{
	//constant pool index to class we are about to create
	u2 iIndex = ReadU2();

	//get class pointer
	cJavaClass *pClass = m_pCurClass->GetClassFromPool(iIndex);
	//allocate memory
	OBJECT *pObj = m_Heap.AllocateClass(pClass);

	//check for out-of-mem
	if(pObj == NULL)
	{
		Error("Error: Out-of-memory.");
		return;
	}

	//push result on top of operand stack
	m_Stack.PushOperandObject(pObj);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_NewArray()
//-----------------------------------------------------------------------------
void cJVM::Inst_NewArray()
{
	//first, get data type
	u1 iType = *m_pPC;
	++m_pPC;

	//get item size
	u2 iItemSize = 0;
	switch(iType)
	{
	case 4: //char
	case 5: //byte
		iItemSize = sizeof(jByte);
		break;
	case 8: //bool
		iItemSize = sizeof(jChar);
		break;
	case 9: //short
		iItemSize = sizeof(jShort);
		break;
	case 10: //int
		iItemSize = sizeof(jInt);
		break;
	case 11: //long
		iItemSize = sizeof(jLong);
		break;
	case 6: //float
		iItemSize = sizeof(jFloat);
		break;
	case 7: //double
		iItemSize = sizeof(jDouble);
		break;
	};

	//get array size
	int iCount = m_Stack.PopOperand().iInt;

	//allocate memory
	OBJECT *pObj = m_Heap.AllocateArray(NULL,iCount,iItemSize);

	//check for out-of-mem
	if(pObj == NULL)
	{
		Error("Error: Out-of-memory.");
		return;
	}

	//push result on top of operand stack
	m_Stack.PushOperandObject(pObj);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ANewArray()
//-----------------------------------------------------------------------------
void cJVM::Inst_ANewArray()
{
	//get constant pool class index
	u2 iIndex = ReadU2();
	//get class info
	cJavaClass *pClass = m_pCurClass->GetClassFromPool(iIndex);

	//get item size
	u2 iItemSize = sizeof(OBJECT*);
	
	//get array size
	int iCount = m_Stack.PopOperand().iInt;

	//allocate memory
	OBJECT *pObj = m_Heap.AllocateArray(pClass,iCount,iItemSize);

	//check for out-of-mem
	if(pObj == NULL)
	{
		Error("Error: Out-of-memory.");
		return;
	}

	//push result on top of operand stack
	m_Stack.PushOperandObject(pObj);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_MultiANewArray()
//-----------------------------------------------------------------------------
void cJVM::Inst_MultiANewArray()
{
	//get constant pool class index
	u2 iIndex = ReadU2();

	u1 iNumDimensions, iItemSize;

	//get array info
	cJavaClass *pClass = m_pCurClass->ParseArrayDefinition(iIndex,iNumDimensions,iItemSize);
	
	//array dimension
	u1 iDim = *m_pPC;
	++m_pPC;

	OBJECT *pObj;
	pObj = CreateMultiDimArray(0,iDim,iItemSize,pClass);
	if(pObj == NULL)
	{
		Error("Error: Out-of-memory.");
		return;
	}

	//pop dimension sizes from stack
	for(u1 i = 0;i < iDim;++i)
	{
		m_Stack.PopOperand();
	}

	//push result on top of operand stack
	m_Stack.PushOperandObject(pObj);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_CheckCast()
//-----------------------------------------------------------------------------
void cJVM::Inst_CheckCast()
{
	//read constant pool index
	u2 iIndex = ReadU2();

	OBJECT *pObj = m_Stack.PopOperand().pObject;
	if(pObj == NULL)
	{
		//return stack to previous state
		m_Stack.PushOperandObject(pObj);
		return;
	}
	if(true == CheckCast(pObj,iIndex))
	{
		//return stack to previous state
		m_Stack.PushOperandObject(pObj);
		return;
	}

	//exception
	Error("Error: ClassCastException.");
	return;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_InstanceOf()
//-----------------------------------------------------------------------------
void cJVM::Inst_InstanceOf()
{
	//read constatn pool index
	u2 iIndex = ReadU2();

	OBJECT *pObj = m_Stack.PopOperand().pObject;

	if(pObj == NULL)
	{
		//
		m_Stack.PushOperandInt(0);
		return;
	}

	if(true == CheckCast(pObj,iIndex))
	{
		m_Stack.PushOperandInt(1);
	}
	else
	{
		m_Stack.PushOperandInt(0);
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_Wide()
//-----------------------------------------------------------------------------
void cJVM::Inst_Wide()
{
	//inst opcode
	u1 opcode = *m_pPC;
	++m_pPC;

	//constatn pool index
	u2 iIndex = ReadU2();

	u4 iHighBytes,iLowBytes;
	jShort iConst;
	STACK_ITEM *pItem;

	switch(opcode)
	{
	case INST_ILOAD:
	case INST_FLOAD:
	case INST_ALOAD:
		m_Stack.PushVariable(iIndex);
		break;

	case INST_ISTORE:
		m_Stack.SetVariableInt(m_Stack.PopOperand().iInt, iIndex);
		break;
	case INST_FSTORE:
		m_Stack.SetVariableFloat(m_Stack.PopOperand().fFloat, iIndex);
		break;
	case INST_ASTORE:
		m_Stack.SetVariableObject(m_Stack.PopOperand().pObject, iIndex);
		break;

	case INST_LLOAD:
	case INST_DLOAD:
		m_Stack.PushVariableDouble(iIndex);
		break;

	case INST_LSTORE:
	case INST_DSTORE:
		//get double
		iLowBytes = m_Stack.PopOperand().iU4;
		iHighBytes = m_Stack.PopOperand().iU4;

		m_Stack.SetVariableU4(iHighBytes,iIndex);
		++iIndex;
		m_Stack.SetVariableU4(iLowBytes,iIndex);
		break;

	case INST_RET:
		pItem = m_Stack.GetVariable(iIndex);

		m_pPC = (u1*)pItem->pPointer;
		break;

	case INST_IINC:
		iConst = (jInt)ReadU2();
		m_Stack.IncVariableInt(iConst,iIndex);
		break;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_Dup()
//-----------------------------------------------------------------------------
void cJVM::Inst_Dup()
{
	//what else to say
	m_Stack.DuplicateTopOperand();
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_Dup_X1()
//-----------------------------------------------------------------------------
void cJVM::Inst_Dup_X1()
{
	m_Stack.DuplicateX1();
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_Dup_X2()
//-----------------------------------------------------------------------------
void cJVM::Inst_Dup_X2()
{
	m_Stack.DuplicateX2();
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_Dup2()
//-----------------------------------------------------------------------------
void cJVM::Inst_Dup2()
{
	//what else to say
	m_Stack.DuplicateTopOperand2();
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_Dup2_X1()
//-----------------------------------------------------------------------------
void cJVM::Inst_Dup2_X1()
{
	m_Stack.Duplicate2X1();
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_Dup2_X2()
//-----------------------------------------------------------------------------
void cJVM::Inst_Dup2_X2()
{
	m_Stack.Duplicate2X2();
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_InvokeInterface()
//-----------------------------------------------------------------------------
void cJVM::Inst_InvokeInterface()
{
	//get constant pool method index
	u2 iIndex = ReadU2();
	u1 iCount = *m_pPC; ++m_pPC;
	++m_pPC;

	//get const pool info
	u1 *pName,*pDesc;
	m_pCurClass->GetInterfaceMethodNameAndDesc(iIndex,&pName,&pDesc);

	//get interface method info
	METHOD_INFO *pMethod = m_pCurClass->GetInterfaceMethodFromPool(iIndex);
	//add one argument for "this" reference
	u2 iNumArgs = pMethod->iNumArguments + 1;

	cJavaClass *pClass = m_Stack.GetOperand(m_Stack.GetOperandStackTop() - iNumArgs).pObject->pClass;
	pMethod = pClass->FindMethod(pName,pDesc);

	//push frame
	m_Stack.PushMethodFrame(pMethod,m_pPC,iNumArgs);

	//update cur class and pc
	m_pCurClass = pMethod->pOwner;
	m_pPC = pMethod->pCode;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_InvokeVirtual()
//-----------------------------------------------------------------------------
void cJVM::Inst_InvokeVirtual()
{
	//get constant pool method index
	u2 iIndex = ReadU2();
	//get method info
	METHOD_INFO *pMethod = m_pCurClass->GetMethodFromPool(iIndex);
	//add one argument for "this" reference
	u2 iNumArgs = pMethod->iNumArguments + 1;

	//push frame
	m_Stack.PushMethodFrame(pMethod,m_pPC,iNumArgs);

	//update cur class and pc
	m_pCurClass = pMethod->pOwner;
	m_pPC = pMethod->pCode;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_InvokeStatic()
//-----------------------------------------------------------------------------
void cJVM::Inst_InvokeStatic()
{
	//get constant pool method index
	u2 iIndex = ReadU2();
	//get method info
	METHOD_INFO *pMethod = m_pCurClass->GetMethodFromPool(iIndex);
	//get num arguments
	u2 iNumArgs = pMethod->iNumArguments;
	//push frame
	m_Stack.PushMethodFrame(pMethod,m_pPC,iNumArgs);

	//update cur class and pc
	m_pCurClass = pMethod->pOwner;
	m_pPC = pMethod->pCode;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_InvokeSpecial()
//-----------------------------------------------------------------------------
void cJVM::Inst_InvokeSpecial()
{
	//get constant pool method index
	u2 iIndex = ReadU2();
	//get method info
	METHOD_INFO *pMethod = m_pCurClass->GetMethodFromPool(iIndex);
	//add one argument for "this" reference
	u2 iNumArgs = pMethod->iNumArguments + 1;
	//push frame
	m_Stack.PushMethodFrame(pMethod,m_pPC,iNumArgs);

	//update cur class and pc
	m_pCurClass = pMethod->pOwner;
	m_pPC = pMethod->pCode;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ALoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_ALoad()
{
	m_Stack.PushVariable(*m_pPC);
	++m_pPC;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_Ret()
//-----------------------------------------------------------------------------
void cJVM::Inst_Ret()
{
	STACK_ITEM *pItem = m_Stack.GetVariable(*m_pPC);
	//++m_pPC;

	m_pPC = (u1*)pItem->pPointer;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LLoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_LLoad()
{
	m_Stack.PushVariableDouble(*m_pPC);
	++m_pPC;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LLoad_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_LLoad_0()
{
	m_Stack.PushVariableDouble(0);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LLoad_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_LLoad_1()
{
	m_Stack.PushVariableDouble(1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LLoad_2()
//-----------------------------------------------------------------------------
void cJVM::Inst_LLoad_2()
{
	m_Stack.PushVariableDouble(2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LLoad_3()
//-----------------------------------------------------------------------------
void cJVM::Inst_LLoad_3()
{
	m_Stack.PushVariableDouble(3);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DLoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_DLoad()
{
	m_Stack.PushVariableDouble(*m_pPC);
	++m_pPC;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DLoad_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_DLoad_0()
{
	m_Stack.PushVariableDouble(0);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DLoad_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_DLoad_1()
{
	m_Stack.PushVariableDouble(1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DLoad_2()
//-----------------------------------------------------------------------------
void cJVM::Inst_DLoad_2()
{
	m_Stack.PushVariableDouble(2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DLoad_3()
//-----------------------------------------------------------------------------
void cJVM::Inst_DLoad_3()
{
	m_Stack.PushVariableDouble(3);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ALoad_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_ALoad_0()
{
	m_Stack.PushVariable(0);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ALoad_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_ALoad_1()
{
	m_Stack.PushVariable(1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ALoad_2()
//-----------------------------------------------------------------------------
void cJVM::Inst_ALoad_2()
{
	m_Stack.PushVariable(2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ALoad_3()
//-----------------------------------------------------------------------------
void cJVM::Inst_ALoad_3()
{
	m_Stack.PushVariable(3);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_Return()
//-----------------------------------------------------------------------------
void cJVM::Inst_Return()
{
	//return from method
	m_pPC = m_Stack.PopMethodFrame(&m_pCurClass);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IReturn()
//-----------------------------------------------------------------------------
void cJVM::Inst_IReturn()
{
	//return and push result to operand stack
	jInt ret = m_Stack.PopOperand().iInt;
	m_pPC = m_Stack.PopMethodFrame(&m_pCurClass);
	m_Stack.PushOperandInt(ret);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FReturn()
//-----------------------------------------------------------------------------
void cJVM::Inst_FReturn()
{
	//return and push result to operand stack
	jFloat ret = m_Stack.PopOperand().fFloat;
	m_pPC = m_Stack.PopMethodFrame(&m_pCurClass);
	m_Stack.PushOperandFloat(ret);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_AReturn()
//-----------------------------------------------------------------------------
void cJVM::Inst_AReturn()
{
	//return and push result to operand stack
	OBJECT *pRet = m_Stack.PopOperand().pObject;
	m_pPC = m_Stack.PopMethodFrame(&m_pCurClass);
	m_Stack.PushOperandObject(pRet);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DReturn()
//-----------------------------------------------------------------------------
void cJVM::Inst_DReturn()
{
	//return and push double to operand stack
	u4 iHighBytes,iLowBytes;
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	m_pPC = m_Stack.PopMethodFrame(&m_pCurClass);

	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LReturn()
//-----------------------------------------------------------------------------
void cJVM::Inst_LReturn()
{
	//return and push double to operand stack
	u4 iHighBytes,iLowBytes;
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	m_pPC = m_Stack.PopMethodFrame(&m_pCurClass);

	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_AStore()
//-----------------------------------------------------------------------------
void cJVM::Inst_AStore()
{
	u1 iIndex = *m_pPC;
	m_pPC++;

	m_Stack.SetVariableObject(m_Stack.PopOperand().pObject, iIndex);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IStore()
//-----------------------------------------------------------------------------
void cJVM::Inst_IStore()
{
	u1 iIndex = *m_pPC;
	m_pPC++;

	m_Stack.SetVariableInt(m_Stack.PopOperand().iInt, iIndex);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FStore()
//-----------------------------------------------------------------------------
void cJVM::Inst_FStore()
{
	u1 iIndex = *m_pPC;
	m_pPC++;

	m_Stack.SetVariableFloat(m_Stack.PopOperand().fFloat, iIndex);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FStore_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_FStore_0()
{
	m_Stack.SetVariableFloat(m_Stack.PopOperand().fFloat, 0);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FStore_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_FStore_1()
{
	m_Stack.SetVariableFloat(m_Stack.PopOperand().fFloat, 1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FStore_2()
//-----------------------------------------------------------------------------
void cJVM::Inst_FStore_2()
{
	m_Stack.SetVariableFloat(m_Stack.PopOperand().fFloat, 2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FStore_3()
//-----------------------------------------------------------------------------
void cJVM::Inst_FStore_3()
{
	m_Stack.SetVariableFloat(m_Stack.PopOperand().fFloat, 3);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DStore()
//-----------------------------------------------------------------------------
void cJVM::Inst_DStore()
{
	//store double from stack to variable
	u1 iIndex = *m_pPC;
	m_pPC++;

	u4 iHighBytes,iLowBytes;
	//get double
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	m_Stack.SetVariableU4(iHighBytes,iIndex);
	++iIndex;
	m_Stack.SetVariableU4(iLowBytes,iIndex);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LStore()
//-----------------------------------------------------------------------------
void cJVM::Inst_LStore()
{
	//store long from stack to variable
	u1 iIndex = *m_pPC;
	m_pPC++;

	u4 iHighBytes,iLowBytes;
	//get double
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	m_Stack.SetVariableU4(iHighBytes,iIndex);
	++iIndex;
	m_Stack.SetVariableU4(iLowBytes,iIndex);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_AStore_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_AStore_0()
{
	m_Stack.SetVariableObject(m_Stack.PopOperand().pObject, 0);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_AStore_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_AStore_1()
{
	m_Stack.SetVariableObject(m_Stack.PopOperand().pObject, 1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_AStore_2()
//-----------------------------------------------------------------------------
void cJVM::Inst_AStore_2()
{
	m_Stack.SetVariableObject(m_Stack.PopOperand().pObject, 2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_AStore_3()
//-----------------------------------------------------------------------------
void cJVM::Inst_AStore_3()
{
	m_Stack.SetVariableObject(m_Stack.PopOperand().pObject, 3);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IStore_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_IStore_0()
{
	m_Stack.SetVariableInt(m_Stack.PopOperand().iInt, 0);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LStore_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_LStore_0()
{
	//double -> store two items
	m_Stack.SetVariableU4(m_Stack.PopOperand().iU4, 1);
	m_Stack.SetVariableU4(m_Stack.PopOperand().iU4, 0);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LStore_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_LStore_1()
{
	//double -> store two items
	m_Stack.SetVariableU4(m_Stack.PopOperand().iU4, 2);
	m_Stack.SetVariableU4(m_Stack.PopOperand().iU4, 1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LStore_2()
//-----------------------------------------------------------------------------
void cJVM::Inst_LStore_2()
{
	//double -> store two items
	m_Stack.SetVariableU4(m_Stack.PopOperand().iU4, 3);
	m_Stack.SetVariableU4(m_Stack.PopOperand().iU4, 2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LStore_3()
//-----------------------------------------------------------------------------
void cJVM::Inst_LStore_3()
{
	//double -> store two items
	m_Stack.SetVariableU4(m_Stack.PopOperand().iU4, 4);
	m_Stack.SetVariableU4(m_Stack.PopOperand().iU4, 3);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DStore_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_DStore_0()
{
	//double -> store two items
	m_Stack.SetVariableU4(m_Stack.PopOperand().iU4, 1);
	m_Stack.SetVariableU4(m_Stack.PopOperand().iU4, 0);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DStore_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_DStore_1()
{
	//double -> store two items
	m_Stack.SetVariableU4(m_Stack.PopOperand().iU4, 2);
	m_Stack.SetVariableU4(m_Stack.PopOperand().iU4, 1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DStore_2()
//-----------------------------------------------------------------------------
void cJVM::Inst_DStore_2()
{
	//double -> store two items
	m_Stack.SetVariableU4(m_Stack.PopOperand().iU4, 3);
	m_Stack.SetVariableU4(m_Stack.PopOperand().iU4, 2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DStore_3()
//-----------------------------------------------------------------------------
void cJVM::Inst_DStore_3()
{
	//double -> store two items
	m_Stack.SetVariableU4(m_Stack.PopOperand().iU4, 4);
	m_Stack.SetVariableU4(m_Stack.PopOperand().iU4, 3);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IStore_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_IStore_1()
{
	m_Stack.SetVariableInt(m_Stack.PopOperand().iInt, 1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IStore_2()
//-----------------------------------------------------------------------------
void cJVM::Inst_IStore_2()
{
	m_Stack.SetVariableInt(m_Stack.PopOperand().iInt, 2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IStore_3()
//-----------------------------------------------------------------------------
void cJVM::Inst_IStore_3()
{
	m_Stack.SetVariableInt(m_Stack.PopOperand().iInt, 3);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FConst_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_FConst_0()
{
	m_Stack.PushOperandFloat(0.0f);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FConst_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_FConst_1()
{
	m_Stack.PushOperandFloat(1.0f);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FConst_2()
//-----------------------------------------------------------------------------
void cJVM::Inst_FConst_2()
{
	m_Stack.PushOperandFloat(2.0f);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IConst_M1()
//-----------------------------------------------------------------------------
void cJVM::Inst_IConst_M1()
{
	m_Stack.PushOperandInt(-1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IConst_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_IConst_0()
{
	m_Stack.PushOperandInt(0);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IConst_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_IConst_1()
{
	m_Stack.PushOperandInt(1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IConst_2()
//-----------------------------------------------------------------------------
void cJVM::Inst_IConst_2()
{
	m_Stack.PushOperandInt(2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IConst_3()
//-----------------------------------------------------------------------------
void cJVM::Inst_IConst_3()
{
	m_Stack.PushOperandInt(3);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IConst_4()
//-----------------------------------------------------------------------------
void cJVM::Inst_IConst_4()
{
	m_Stack.PushOperandInt(4);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IConst_5()
//-----------------------------------------------------------------------------
void cJVM::Inst_IConst_5()
{
	m_Stack.PushOperandInt(5);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ILoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_ILoad()
{
	u1 iIndex = *m_pPC;
	m_pPC++;

	m_Stack.PushVariable(iIndex);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FLoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_FLoad()
{
	u1 iIndex = *m_pPC;
	m_pPC++;

	m_Stack.PushVariable(iIndex);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FLoad_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_FLoad_0()
{
	m_Stack.PushVariable(0);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FLoad_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_FLoad_1()
{
	m_Stack.PushVariable(1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FLoad_2()
//-----------------------------------------------------------------------------
void cJVM::Inst_FLoad_2()
{
	m_Stack.PushVariable(2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FLoad_3()
//-----------------------------------------------------------------------------
void cJVM::Inst_FLoad_3()
{
	m_Stack.PushVariable(3);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ILoad_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_ILoad_0()
{
	m_Stack.PushVariable(0);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ILoad_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_ILoad_1()
{
	m_Stack.PushVariable(1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ILoad_2()
//-----------------------------------------------------------------------------
void cJVM::Inst_ILoad_2()
{
	m_Stack.PushVariable(2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ILoad_3()
//-----------------------------------------------------------------------------
void cJVM::Inst_ILoad_3()
{
	m_Stack.PushVariable(3);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_AConst_Null()
//-----------------------------------------------------------------------------
void cJVM::Inst_AConst_Null()
{
	m_Stack.PushOperandObject(NULL);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IAdd()
//-----------------------------------------------------------------------------
void cJVM::Inst_IAdd()
{
	m_Stack.PushOperandInt(m_Stack.PopOperand().iInt + m_Stack.PopOperand().iInt);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IAnd()
//-----------------------------------------------------------------------------
void cJVM::Inst_IAnd()
{
	jInt i2 = m_Stack.PopOperand().iInt;
	jInt i1 = m_Stack.PopOperand().iInt;

	jInt res = i1 & i2;
	m_Stack.PushOperandInt(res);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IDiv()
//-----------------------------------------------------------------------------
void cJVM::Inst_IDiv()
{
	jInt i2 = m_Stack.PopOperand().iInt;
	jInt i1 = m_Stack.PopOperand().iInt;

	jInt res = i1 / i2;
	m_Stack.PushOperandInt(res);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IOr()
//-----------------------------------------------------------------------------
void cJVM::Inst_IOr()
{
	jInt i2 = m_Stack.PopOperand().iInt;
	jInt i1 = m_Stack.PopOperand().iInt;

	jInt res = i1 | i2;
	m_Stack.PushOperandInt(res);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IXOr()
//-----------------------------------------------------------------------------
void cJVM::Inst_IXOr()
{
	jInt i2 = m_Stack.PopOperand().iInt;
	jInt i1 = m_Stack.PopOperand().iInt;

	jInt res = i1 ^ i2;
	m_Stack.PushOperandInt(res);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IRem()
//-----------------------------------------------------------------------------
void cJVM::Inst_IRem()
{
	jInt i2 = m_Stack.PopOperand().iInt;
	jInt i1 = m_Stack.PopOperand().iInt;

	jInt res = i1 % i2;
	m_Stack.PushOperandInt(res);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ISHL()
//-----------------------------------------------------------------------------
void cJVM::Inst_ISHL()
{
	jInt i2 = m_Stack.PopOperand().iInt;
	jInt i1 = m_Stack.PopOperand().iInt;

	jInt res = i1 << i2;
	m_Stack.PushOperandInt(res);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ISHR()
//-----------------------------------------------------------------------------
void cJVM::Inst_ISHR()
{
	jInt i2 = m_Stack.PopOperand().iInt;
	jInt i1 = m_Stack.PopOperand().iInt;

	jInt res = i1 >> i2;
	m_Stack.PushOperandInt(res);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IUSHR()
//-----------------------------------------------------------------------------
void cJVM::Inst_IUSHR()
{
	u4 i2 = m_Stack.PopOperand().iInt;
	jInt i1 = m_Stack.PopOperand().iInt;

	jInt res = i1 >> i2;
	m_Stack.PushOperandInt(res);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IMul()
//-----------------------------------------------------------------------------
void cJVM::Inst_IMul()
{
	jInt i2 = m_Stack.PopOperand().iInt;
	jInt i1 = m_Stack.PopOperand().iInt;

	jInt res = i1 * i2;
	m_Stack.PushOperandInt(res);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ISub()
//-----------------------------------------------------------------------------
void cJVM::Inst_ISub()
{
	jInt i2 = m_Stack.PopOperand().iInt;
	jInt i1 = m_Stack.PopOperand().iInt;

	jInt res = i1 - i2;
	m_Stack.PushOperandInt(res);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_INeg()
//-----------------------------------------------------------------------------
void cJVM::Inst_INeg()
{
	jInt i = m_Stack.PopOperand().iInt;

	i = -i;
	m_Stack.PushOperandInt(i);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FAdd()
//-----------------------------------------------------------------------------
void cJVM::Inst_FAdd()
{
	m_Stack.PushOperandFloat(m_Stack.PopOperand().fFloat + m_Stack.PopOperand().fFloat);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FSub()
//-----------------------------------------------------------------------------
void cJVM::Inst_FSub()
{
	jFloat value2 = m_Stack.PopOperand().fFloat;
	jFloat value1 = m_Stack.PopOperand().fFloat;

	m_Stack.PushOperandFloat(value1 - value2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FDiv()
//-----------------------------------------------------------------------------
void cJVM::Inst_FDiv()
{
	jFloat value2 = m_Stack.PopOperand().fFloat;
	jFloat value1 = m_Stack.PopOperand().fFloat;

	m_Stack.PushOperandFloat(value1 / value2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FRem()
//-----------------------------------------------------------------------------
void cJVM::Inst_FRem()
{
	jFloat value2 = m_Stack.PopOperand().fFloat;
	jFloat value1 = m_Stack.PopOperand().fFloat;

	m_Stack.PushOperandFloat(fmod(value1,value2));
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FMul()
//-----------------------------------------------------------------------------
void cJVM::Inst_FMul()
{
	jFloat value2 = m_Stack.PopOperand().fFloat;
	jFloat value1 = m_Stack.PopOperand().fFloat;

	m_Stack.PushOperandFloat(value1 * value2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FNeg()
//-----------------------------------------------------------------------------
void cJVM::Inst_FNeg()
{
	jFloat value = m_Stack.PopOperand().fFloat;

	m_Stack.PushOperandFloat(-value);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_Pop()
//-----------------------------------------------------------------------------
void cJVM::Inst_Pop()
{
	m_Stack.PopOperand();
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_Pop2()
//-----------------------------------------------------------------------------
void cJVM::Inst_Pop2()
{
	m_Stack.PopOperand();
	m_Stack.PopOperand();
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IAStore()
//-----------------------------------------------------------------------------
void cJVM::Inst_IAStore()
{
	//get value
	jInt iValue = m_Stack.PopOperand().iInt;
	//get array index
	jInt iIndex = m_Stack.PopOperand().iInt;
	//get array
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (jInt)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//get size of one item in array
	jInt iItemSize = (pObj->iSize - sizeof(OBJECT)) / pObj->iArraySize;
	//skip object header
	u1 *pArray = (u1*)(pObj + 1);
	//move to exact position 
	pArray = pArray + (iIndex * iItemSize);
	//copy value
	memcpy(pArray,&iValue,sizeof(jInt));
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_SAStore()
//-----------------------------------------------------------------------------
void cJVM::Inst_SAStore()
{
	//get value
	jShort sValue = m_Stack.PopOperand().iInt;
	//get array index
	jInt iIndex = m_Stack.PopOperand().iInt;
	//get array
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (jInt)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//get size of one item in array
	jInt iItemSize = (pObj->iSize - sizeof(OBJECT)) / pObj->iArraySize;
	//skip object header
	u1 *pArray = (u1*)(pObj + 1);
	//move to exact position 
	pArray = pArray + (iIndex * iItemSize);
	//copy value
	memcpy(pArray,&sValue,sizeof(jShort));
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FAStore()
//-----------------------------------------------------------------------------
void cJVM::Inst_FAStore()
{
	//get value
	jFloat fValue = m_Stack.PopOperand().fFloat;
	//get array index
	jInt iIndex = m_Stack.PopOperand().iInt;
	//get array
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (jInt)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//get size of one item in array
	jInt iItemSize = (pObj->iSize - sizeof(OBJECT)) / pObj->iArraySize;
	//skip object header
	u1 *pArray = (u1*)(pObj + 1);
	//move to exact position 
	pArray = pArray + (iIndex * iItemSize);
	//copy value
	memcpy(pArray,&fValue,sizeof(jFloat));
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DAStore()
//-----------------------------------------------------------------------------
void cJVM::Inst_DAStore()
{
	//get value
	u4 iLow = (u4)m_Stack.PopOperand().iInt;
	u4 iHigh = (u4)m_Stack.PopOperand().iInt;
	jDouble d;
	this->GetDouble(d,iHigh,iLow);
	//get array index
	jInt iIndex = m_Stack.PopOperand().iInt;
	//get array
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (jInt)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//get size of one item in array
	jInt iItemSize = (pObj->iSize - sizeof(OBJECT)) / pObj->iArraySize;
	//skip object header
	u1 *pArray = (u1*)(pObj + 1);
	//move to exact position 
	pArray = pArray + (iIndex * iItemSize);
	//copy value
	memcpy(pArray,&d,sizeof(jDouble));
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LAStore()
//-----------------------------------------------------------------------------
void cJVM::Inst_LAStore()
{
	//get value
	u4 iLow = m_Stack.PopOperand().iU4;
	u4 iHigh = m_Stack.PopOperand().iU4;
	jLong l;
	this->GetLong(l,iHigh,iLow);
	//get array index
	jInt iIndex = m_Stack.PopOperand().iInt;
	//get array
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (jInt)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//get size of one item in array
	jInt iItemSize = (pObj->iSize - sizeof(OBJECT)) / pObj->iArraySize;
	//skip object header
	u1 *pArray = (u1*)(pObj + 1);
	//move to exact position 
	pArray = pArray + (iIndex * iItemSize);
	//copy value
	memcpy(pArray,&l,sizeof(jLong));
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_BIPush()
//-----------------------------------------------------------------------------
void cJVM::Inst_BIPush()
{
	jInt iValue = (jInt)(*m_pPC);
	++m_pPC;
	m_Stack.PushOperandInt(iValue);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_PutStatic()
//-----------------------------------------------------------------------------
void cJVM::Inst_PutStatic()
{
	//write static
	//get constant pool index of field
	u2 iIndex = ReadU2();
	//get field info
	FIELD_INFO *pField = m_pCurClass->GetFieldFromPool(iIndex);

	//determine field type
	if(pField->iSize == sizeof(jDouble))
	{
		jInt i = m_Stack.PopOperand().iInt;
		//low bytes
		pField->pOwner->SetStaticValue(pField->iOffset + sizeof(jDouble)/2,&i,sizeof(jDouble)/2);
		i = m_Stack.PopOperand().iInt;
		//high bytes
		pField->pOwner->SetStaticValue(pField->iOffset,&i,sizeof(jDouble)/2);
		
	}
	else if(pField->iSize == sizeof(jShort)) //short, char
	{
		jShort s = (jShort)m_Stack.PopOperand().iInt;			
		pField->pOwner->SetStaticValue(pField->iOffset,&s,pField->iSize);
	}
	else if(pField->iSize == sizeof(jByte)) //bool, byte
	{
		jByte c = (jByte)m_Stack.PopOperand().iInt;
		pField->pOwner->SetStaticValue(pField->iOffset,&c,pField->iSize);
	}
	else //everything else is 4-bytes
	{
		jInt i = m_Stack.PopOperand().iInt;
		pField->pOwner->SetStaticValue(pField->iOffset,&i,pField->iSize);
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_GetStatic()
//-----------------------------------------------------------------------------
void cJVM::Inst_GetStatic()
{
	//load static
	//get constant pool index of field
	u2 iIndex = ReadU2();
	//get field info
	FIELD_INFO *pField = m_pCurClass->GetFieldFromPool(iIndex);
	//determine field size
	if(pField->iSize == sizeof(jDouble))
	{
		u4 i;			
		//high bytes
		pField->pOwner->GetStaticValue(pField->iOffset,&i,sizeof(jDouble)/2);
		m_Stack.PushOperandU4(i);
		//low bytes
		pField->pOwner->GetStaticValue(pField->iOffset + sizeof(jDouble)/2,&i,sizeof(jDouble)/2);
		m_Stack.PushOperandU4(i);
	}
	else if(pField->iSize == sizeof(jShort)) //short, char
	{
		jShort s;
		pField->pOwner->GetStaticValue(pField->iOffset,&s,pField->iSize);
		m_Stack.PushOperandInt((jInt)s);
	}
	else if(pField->iSize == sizeof(jByte)) // bool, byte
	{
		jByte c;
		pField->pOwner->GetStaticValue(pField->iOffset,&c,pField->iSize);
		m_Stack.PushOperandInt((jInt)c);
	}
	else //everything else is four byte
	{
		STACK_ITEM i;
		pField->pOwner->GetStaticValue(pField->iOffset,&i,pField->iSize);
		m_Stack.PushOperand(i,pField->bReference); 
	}

}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_PutField()
//-----------------------------------------------------------------------------
void cJVM::Inst_PutField()
{
	//get field info from constant pool
	u2 iIndex = ReadU2();
	FIELD_INFO *pField = m_pCurClass->GetFieldFromPool(iIndex);

	//get value to put
	jInt item1 = m_Stack.PopOperand().iInt;
	jInt item2;
	//if field is double or long, fetch second part of value
	if(pField->iSize == sizeof(jDouble))
		item2 = m_Stack.PopOperand().iInt;

	//now, get object
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}

	//skip object header
	u1 *pData = (u1*)(pObj + 1);

	//based on field data type copy value
	if(pField->iSize == sizeof(jDouble))
	{ //double, long
		//low bytes
		memcpy(&pData[pField->iOffset + sizeof(jDouble)/2], &item1, sizeof(jDouble)/2);
		//high bytes
		memcpy(&pData[pField->iOffset], &item2, sizeof(jDouble)/2);
	}
	else if(pField->iSize == sizeof(jShort))
	{
		jShort s = (jShort)item1;
		memcpy(&pData[pField->iOffset], &s, pField->iSize);
	}
	else if(pField->iSize == sizeof(jByte))
	{
		jByte c = (jByte)item1;
		memcpy(&pData[pField->iOffset], &c, pField->iSize);
	}
	else //everything else if four-byte
	{
		memcpy(&pData[pField->iOffset], &item1, pField->iSize);
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_GetField()
//-----------------------------------------------------------------------------
void cJVM::Inst_GetField()
{
	//get field info
	u2 iIndex = ReadU2();
	FIELD_INFO *pField = m_pCurClass->GetFieldFromPool(iIndex);
	//get object
	OBJECT *pObj = m_Stack.PopOperand().pObject;
	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}

	//skip object header
	u1 *pData = (u1*)(pObj + 1);

	STACK_ITEM item;
	//based on field data type fetch value
	if(pField->iSize == sizeof(jDouble))
	{			
		//high bytes
		memcpy(&item, &pData[pField->iOffset], sizeof(jDouble)/2);
		m_Stack.PushOperand(item,0);

		//low bytes
		memcpy(&item, &pData[pField->iOffset + sizeof(jDouble)/2], sizeof(jDouble)/2);
		m_Stack.PushOperand(item,0);
	}
	else if(pField->iSize == sizeof(jShort))
	{
		jShort s = (jShort)pData[pField->iOffset];
		m_Stack.PushOperandInt((jInt)s);
	}
	else if(pField->iSize == sizeof(jByte))
	{
		jByte c = (jByte)pData[pField->iOffset];
		m_Stack.PushOperandInt((jInt)c);
	}
	else //everything else is 4-byte
	{
		memcpy(&item, &pData[pField->iOffset], pField->iSize);
		m_Stack.PushOperand(item,pField->bReference);
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_GoTo()
//-----------------------------------------------------------------------------
void cJVM::Inst_GoTo()
{
	u2 iOffset = (u2)ReadU2();
	//offset is from beginnig of instruction, so -3
	m_pPC += iOffset - 3;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_GoTo_W()
//-----------------------------------------------------------------------------
void cJVM::Inst_GoTo_W()
{
	u4 iOffset = (u4)ReadU4();
	//offset is from beginnig of instruction, so -5
	m_pPC += iOffset - 5;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_I2B()
//-----------------------------------------------------------------------------
void cJVM::Inst_I2B()
{
	jByte byte = (jByte)m_Stack.PopOperand().iInt;

	m_Stack.PushOperandInt((jInt)byte);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_I2C()
//-----------------------------------------------------------------------------
void cJVM::Inst_I2C()
{
	jChar c = (jChar)m_Stack.PopOperand().iInt;

	m_Stack.PushOperandInt((jInt)c);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_I2D()
//-----------------------------------------------------------------------------
void cJVM::Inst_I2D()
{
	jDouble d = (jDouble)m_Stack.PopOperand().iInt;

	u4 iHigh,iLow;
	GetBytesFromDouble(d,iHigh,iLow);

	m_Stack.PushOperandInt(iHigh);
	m_Stack.PushOperandInt(iLow);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_I2L()
//-----------------------------------------------------------------------------
void cJVM::Inst_I2L()
{
	jLong l = (jLong)m_Stack.PopOperand().iInt;

	u4 iHigh,iLow;
	this->GetBytesFromLong(l,iHigh,iLow);

	m_Stack.PushOperandInt(iHigh);
	m_Stack.PushOperandInt(iLow);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_I2F()
//-----------------------------------------------------------------------------
void cJVM::Inst_I2F()
{
	jFloat f = (jFloat)m_Stack.PopOperand().iInt;

	m_Stack.PushOperandFloat(f);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_I2S()
//-----------------------------------------------------------------------------
void cJVM::Inst_I2S()
{
	jShort s = (jShort)m_Stack.PopOperand().iInt;

	m_Stack.PushOperandInt((jInt)s);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_JSR()
//-----------------------------------------------------------------------------
void cJVM::Inst_JSR()
{
	u2 iOffset = ReadU2();

	//push return address on top of the stack
	STACK_ITEM item;
	item.pPointer = m_pPC;
	m_Stack.PushOperand(item,false);

	m_pPC += iOffset - 3;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_JSR_W()
//-----------------------------------------------------------------------------
void cJVM::Inst_JSR_W()
{
	u4 iOffset = ReadU4();

	//push return address on top of the stack
	STACK_ITEM item;
	item.pPointer = m_pPC;
	m_Stack.PushOperand(item,false);

	m_pPC += iOffset - 5;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_BALoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_BALoad()
{
	//get array index
	jInt iIndex = m_Stack.PopOperand().iInt;
	//get object
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (jInt)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//skip object header
	u1 *pData = (u1*)(pObj + 1);

	//push array value
	jByte b = ((jByte*)pData)[iIndex];
	m_Stack.PushOperandInt((jInt)b);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_CALoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_CALoad()
{
	//get array index
	jInt iIndex = m_Stack.PopOperand().iInt;
	//get object
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (jInt)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//skip object header
	u1 *pData = (u1*)(pObj + 1);

	//push array value
	jChar c = ((jChar*)pData)[iIndex];
	m_Stack.PushOperandInt(c);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IInc()
//-----------------------------------------------------------------------------
void cJVM::Inst_IInc()
{
	//increament variable
	u1 iIndex = *m_pPC;
	++m_pPC;
	jInt iConst = *m_pPC;
	++m_pPC;

	m_Stack.IncVariableInt(iConst,iIndex);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ArrayLength()
//-----------------------------------------------------------------------------
void cJVM::Inst_ArrayLength()
{
	//get array header
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}

	//push array size
	m_Stack.PushOperandU4(pObj->iArraySize);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_If_ICMPEQ()
//-----------------------------------------------------------------------------
void cJVM::Inst_If_ICMPEQ()
{
	//get jump target (pc offset)
	jShort iOffset = (jShort)ReadU2();

	//pop values
	jInt iValue2 = m_Stack.PopOperand().iInt;
	jInt iValue1 = m_Stack.PopOperand().iInt;

	//if true
	if(iValue1 == iValue2)
	{
		//jump (offset from beginning of instruction -> -3)
		m_pPC += iOffset - 3;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_If_ICMPNE()
//-----------------------------------------------------------------------------
void cJVM::Inst_If_ICMPNE()
{
	//get jump target (pc offset)
	jShort iOffset = (jShort)ReadU2();

	//pop values
	jInt iValue2 = m_Stack.PopOperand().iInt;
	jInt iValue1 = m_Stack.PopOperand().iInt;

	//if true
	if(iValue1 != iValue2)
	{
		//jump (offset from beginning of instruction -> -3)
		m_pPC += iOffset - 3;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_If_ICMPLT()
//-----------------------------------------------------------------------------
void cJVM::Inst_If_ICMPLT()
{
	//get jump target (pc offset)
	jShort iOffset = (jShort)ReadU2();

	//pop values
	jInt iValue2 = m_Stack.PopOperand().iInt;
	jInt iValue1 = m_Stack.PopOperand().iInt;

	//if true
	if(iValue1 < iValue2)
	{
		//jump (offset from beginning of instruction -> -3)
		m_pPC += iOffset - 3;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_If_ICMPLE()
//-----------------------------------------------------------------------------
void cJVM::Inst_If_ICMPLE()
{
	//get jump target (pc offset)
	jShort iOffset = (jShort)ReadU2();

	//pop values
	jInt iValue2 = m_Stack.PopOperand().iInt;
	jInt iValue1 = m_Stack.PopOperand().iInt;

	//if true
	if(iValue1 <= iValue2)
	{
		//jump (offset from beginning of instruction -> -3)
		m_pPC += iOffset - 3;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_If_ICMPGT()
//-----------------------------------------------------------------------------
void cJVM::Inst_If_ICMPGT()
{
	//get jump target (pc offset)
	jShort iOffset = (jShort)ReadU2();

	//pop values
	jInt iValue2 = m_Stack.PopOperand().iInt;
	jInt iValue1 = m_Stack.PopOperand().iInt;

	//if true
	if(iValue1 > iValue2)
	{
		//jump (offset from beginning of instruction -> -3)
		m_pPC += iOffset - 3;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_If_ICMPGE()
//-----------------------------------------------------------------------------
void cJVM::Inst_If_ICMPGE()
{
	//get jump target (pc offset)
	jShort iOffset = (jShort)ReadU2();

	//pop values
	jInt iValue2 = m_Stack.PopOperand().iInt;
	jInt iValue1 = m_Stack.PopOperand().iInt;

	//if true
	if(iValue1 >= iValue2)
	{
		//jump (offset from beginning of instruction -> -3)
		m_pPC += iOffset - 3;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_If_ACMPEQ()
//-----------------------------------------------------------------------------
void cJVM::Inst_If_ACMPEQ()
{
	//get jump target (pc offset)
	jShort iOffset = (jShort)ReadU2();

	//pop values
	OBJECT *pRef2 = m_Stack.PopOperand().pObject;
	OBJECT *pRef1 = m_Stack.PopOperand().pObject;

	//if true
	if(pRef1 == pRef2)
	{
		//jump (offset from beginning of instruction -> -3)
		m_pPC += iOffset - 3;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_If_ACMPNE()
//-----------------------------------------------------------------------------
void cJVM::Inst_If_ACMPNE()
{
	//get jump target (pc offset)
	jShort iOffset = (jShort)ReadU2();

	//pop values
	OBJECT *pRef2 = m_Stack.PopOperand().pObject;
	OBJECT *pRef1 = m_Stack.PopOperand().pObject;

	//if true
	if(pRef1 != pRef2)
	{
		//jump (offset from beginning of instruction -> -3)
		m_pPC += iOffset - 3;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IFEQ()
//-----------------------------------------------------------------------------
void cJVM::Inst_IFEQ()
{
	//get jump target (pc offset)
	jShort iOffset = (jShort)ReadU2();

	//pop values
	jInt iValue = m_Stack.PopOperand().iInt;

	//if true
	if(iValue == 0)
	{
		//jump (offset from beginning of instruction -> -3)
		m_pPC += iOffset - 3;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IFNE()
//-----------------------------------------------------------------------------
void cJVM::Inst_IFNE()
{
	//get jump target (pc offset)
	jShort iOffset = (jShort)ReadU2();

	//pop values
	jInt iValue = m_Stack.PopOperand().iInt;

	//if true
	if(iValue != 0)
	{
		//jump (offset from beginning of instruction -> -3)
		m_pPC += iOffset - 3;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IFLT()
//-----------------------------------------------------------------------------
void cJVM::Inst_IFLT()
{
	//get jump target (pc offset)
	jShort iOffset = (jShort)ReadU2();

	//pop values
	jInt iValue = m_Stack.PopOperand().iInt;

	//if true
	if(iValue < 0)
	{
		//jump (offset from beginning of instruction -> -3)
		m_pPC += iOffset - 3;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IFLE()
//-----------------------------------------------------------------------------
void cJVM::Inst_IFLE()
{
	//get jump target (pc offset)
	jShort iOffset = (jShort)ReadU2();

	//pop values
	jInt iValue = m_Stack.PopOperand().iInt;

	//if true
	if(iValue <= 0)
	{
		//jump (offset from beginning of instruction -> -3)
		m_pPC += iOffset - 3;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IFGT()
//-----------------------------------------------------------------------------
void cJVM::Inst_IFGT()
{
	//get jump target (pc offset)
	jShort iOffset = (jShort)ReadU2();

	//pop values
	jInt iValue = m_Stack.PopOperand().iInt;

	//if true
	if(iValue > 0)
	{
		//jump (offset from beginning of instruction -> -3)
		m_pPC += iOffset - 3;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IFGE()
//-----------------------------------------------------------------------------
void cJVM::Inst_IFGE()
{
	//get jump target (pc offset)
	jShort iOffset = (jShort)ReadU2();

	//pop values
	jInt iValue = m_Stack.PopOperand().iInt;

	//if true
	if(iValue >= 0)
	{
		//jump (offset from beginning of instruction -> -3)
		m_pPC += iOffset - 3;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IfNonNull()
//-----------------------------------------------------------------------------
void cJVM::Inst_IfNonNull()
{
	//get jump target (pc offset)
	jShort iOffset = (jShort)ReadU2();

	//pop values
	OBJECT *pValue = m_Stack.PopOperand().pObject;

	//if true
	if(pValue != NULL)
	{
		//jump (offset from beginning of instruction -> -3)
		m_pPC += iOffset - 3;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IfNull()
//-----------------------------------------------------------------------------
void cJVM::Inst_IfNull()
{
	//get jump target (pc offset)
	jShort iOffset = (jShort)ReadU2();

	//pop values
	OBJECT *pValue = m_Stack.PopOperand().pObject;

	//if true
	if(pValue == NULL)
	{
		//jump (offset from beginning of instruction -> -3)
		m_pPC += iOffset - 3;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_SIPush()
//-----------------------------------------------------------------------------
void cJVM::Inst_SIPush()
{
	//push short
	jShort s = (jShort)ReadU2();
	m_Stack.PushOperandInt(s);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_Swap()
//-----------------------------------------------------------------------------
void cJVM::Inst_Swap()
{
	m_Stack.SwapTopOperands();
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_BAStore()
//-----------------------------------------------------------------------------
void cJVM::Inst_BAStore()
{
	//get value and array index
	jByte iValue = m_Stack.PopOperand().iInt;
	jInt iIndex = m_Stack.PopOperand().iInt;

	//get object
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (jInt)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//skip object's header
	jByte *pData = (jByte*)(pObj + 1);

	//write byte
	memcpy(&pData[iIndex],&iValue,sizeof(jByte));
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_CAStore()
//-----------------------------------------------------------------------------
void cJVM::Inst_CAStore()
{
	//get value and array index
	jChar iValue = m_Stack.PopOperand().iInt;
	jInt iIndex = m_Stack.PopOperand().iInt;

	//get object
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (jInt)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//skip object's header
	jChar *pData = (jChar*)(pObj + 1);

	//write byte
	memcpy(&pData[iIndex],&iValue,sizeof(jChar));
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IALoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_IALoad()
{
	//get inex and object
	jInt iIndex = m_Stack.PopOperand().iInt;
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (jInt)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//skip header
	u1 *pData = (u1*)(pObj + 1);

	jInt i;
	//copy value
	memcpy(&i,&pData[iIndex * sizeof(jInt)],sizeof(jInt));
	//push result
	m_Stack.PushOperandInt(i);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_SALoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_SALoad()
{
	//get inex and object
	jInt iIndex = m_Stack.PopOperand().iInt;
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (jInt)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//skip header
	u1 *pData = (u1*)(pObj + 1);

	jShort s;
	//copy value
	memcpy(&s,&pData[iIndex * sizeof(jShort)],sizeof(jShort));
	//push result
	m_Stack.PushOperandInt(s);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FALoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_FALoad()
{
	//get inex and object
	jInt iIndex = m_Stack.PopOperand().iInt;
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (jInt)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//skip header
	u1 *pData = (u1*)(pObj + 1);

	jFloat f;
	//copy value
	memcpy(&f,&pData[iIndex * sizeof(jFloat)],sizeof(jFloat));
	//push result
	m_Stack.PushOperandFloat(f);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DALoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_DALoad()
{
	//get inex and object
	jInt iIndex = m_Stack.PopOperand().iInt;
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (jInt)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//skip header
	u1 *pData = (u1*)(pObj + 1);

	jDouble d;
	//copy value
	memcpy(&d,&pData[iIndex * sizeof(jDouble)],sizeof(jDouble));

	u4 iHigh, iLow;
	this->GetBytesFromDouble(d,iHigh,iLow);

	//push result
	m_Stack.PushOperandU4(iHigh);
	m_Stack.PushOperandU4(iLow);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LALoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_LALoad()
{
	//get inex and object
	jInt iIndex = m_Stack.PopOperand().iInt;
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (jInt)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//skip header
	u1 *pData = (u1*)(pObj + 1);

	jLong l;
	//copy value
	memcpy(&l,&pData[iIndex * sizeof(jLong)],sizeof(jLong));

	u4 iHigh, iLow;
	this->GetBytesFromLong(l,iHigh,iLow);

	//push result
	m_Stack.PushOperandU4(iHigh);
	m_Stack.PushOperandU4(iLow);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_AALoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_AALoad()
{
	//get index and object
	jInt iIndex = m_Stack.PopOperand().iInt;
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (jInt)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//skip header
	OBJECT **ppData = (OBJECT**)(pObj + 1);

	//push result
	m_Stack.PushOperandObject(ppData[iIndex]);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LDC2_W()
//-----------------------------------------------------------------------------
void cJVM::Inst_LDC2_W()
{
	//load double(long) from constant pool

	//constant pool index
	u2 iIndex = ReadU2();
	u4 iHighBytes, iLowBytes;
	
	//get value
	m_pCurClass->GetDoubleOrLongFromConstantPool(iIndex,iHighBytes,iLowBytes);
	//push both parts
	m_Stack.PushOperandInt(iHighBytes);
	m_Stack.PushOperandInt(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LDC()
//-----------------------------------------------------------------------------
void cJVM::Inst_LDC()
{
	//load float(int) from constant pool

	//constant pool index
	u1 iIndex = *m_pPC;
	++m_pPC;

	//get const
	u4 constant;
	m_pCurClass->GetFloatOrIntFromConstantPool(iIndex,&constant);
	
	//push const
	m_Stack.PushOperandU4(constant);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LDC_W()
//-----------------------------------------------------------------------------
void cJVM::Inst_LDC_W()
{
	//load float(int) from constant pool

	//constant pool index
	u2 iIndex = ReadU2();

	//get const
	u4 constant;
	m_pCurClass->GetFloatOrIntFromConstantPool(iIndex,&constant);
	
	//push const
	m_Stack.PushOperandU4(constant);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LAdd()
//-----------------------------------------------------------------------------
void cJVM::Inst_LAdd()
{
	jLong l1,l2;
	u4 iHighBytes,iLowBytes;
	//get first double
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	//assembly first double
	this->GetLong(l2,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	this->GetLong(l1,iHighBytes,iLowBytes);

	//add
	jLong l3 = l1 + l2;
	
	//disassembly result
	this->GetBytesFromLong(l3,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LDiv()
//-----------------------------------------------------------------------------
void cJVM::Inst_LDiv()
{
	jLong l1,l2;
	u4 iHighBytes,iLowBytes;
	//get first double
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	//assembly first double
	this->GetLong(l2,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	this->GetLong(l1,iHighBytes,iLowBytes);

	//div
	jLong l3 = l1 / l2;
	
	//disassembly result
	this->GetBytesFromLong(l3,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LSub()
//-----------------------------------------------------------------------------
void cJVM::Inst_LSub()
{
	jLong l1,l2;
	u4 iHighBytes,iLowBytes;
	//get first double
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	//assembly first double
	this->GetLong(l2,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	this->GetLong(l1,iHighBytes,iLowBytes);

	//div
	jLong l3 = l1 - l2;
	
	//disassembly result
	this->GetBytesFromLong(l3,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LShl()
//-----------------------------------------------------------------------------
void cJVM::Inst_LShl()
{
	jLong l1,l2;
	u4 iHighBytes,iLowBytes;
	//get first long
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	//assembly first long
	this->GetLong(l2,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	this->GetLong(l1,iHighBytes,iLowBytes);

	//shl
	jLong l3 = l1 << l2;
	
	//disassembly result
	this->GetBytesFromLong(l3,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LShr()
//-----------------------------------------------------------------------------
void cJVM::Inst_LShr()
{
	jLong l1,l2;
	u4 iHighBytes,iLowBytes;
	//get first double
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	//assembly first double
	this->GetLong(l2,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	this->GetLong(l1,iHighBytes,iLowBytes);

	//div
	jLong l3 = l1 >> l2;
	
	//disassembly result
	this->GetBytesFromLong(l3,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LUShr()
//-----------------------------------------------------------------------------
void cJVM::Inst_LUShr()
{
	jLong l1,l2;
	u4 iHighBytes,iLowBytes;
	//get first double
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	//assembly first double
	this->GetLong(l2,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	this->GetLong(l1,iHighBytes,iLowBytes);

	//div
	jLong l3 = (jULong)l1 >> l2;
	
	//disassembly result
	this->GetBytesFromLong(l3,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LMul()
//-----------------------------------------------------------------------------
void cJVM::Inst_LMul()
{
	jLong l1,l2;
	u4 iHighBytes,iLowBytes;
	//get first double
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	//assembly first double
	this->GetLong(l2,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	this->GetLong(l1,iHighBytes,iLowBytes);

	//mul
	jLong l3 = l1 * l2;
	
	//disassembly result
	this->GetBytesFromLong(l3,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LOr()
//-----------------------------------------------------------------------------
void cJVM::Inst_LOr()
{
	jLong l1,l2;
	u4 iHighBytes,iLowBytes;
	//get first double
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	//assembly first double
	this->GetLong(l2,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	this->GetLong(l1,iHighBytes,iLowBytes);

	//or
	jLong l3 = l1 | l2;
	
	//disassembly result
	this->GetBytesFromLong(l3,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LXOr()
//-----------------------------------------------------------------------------
void cJVM::Inst_LXOr()
{
	jLong l1,l2;
	u4 iHighBytes,iLowBytes;
	//get first double
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	//assembly first double
	this->GetLong(l2,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	this->GetLong(l1,iHighBytes,iLowBytes);

	//or
	jLong l3 = l1 ^ l2;
	
	//disassembly result
	this->GetBytesFromLong(l3,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LRem()
//-----------------------------------------------------------------------------
void cJVM::Inst_LRem()
{
	jLong l1,l2;
	u4 iHighBytes,iLowBytes;
	//get first double
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	//assembly first double
	this->GetLong(l2,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	this->GetLong(l1,iHighBytes,iLowBytes);

	//rem
	jLong l3 = l1 % l2;
	
	//disassembly result
	this->GetBytesFromLong(l3,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LNeg()
//-----------------------------------------------------------------------------
void cJVM::Inst_LNeg()
{
	jLong l;
	u4 iHighBytes,iLowBytes;
	//get double
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	//assembly first double
	this->GetLong(l,iHighBytes,iLowBytes);

	l = -l;
	
	//disassembly result
	this->GetBytesFromLong(l,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LAnd()
//-----------------------------------------------------------------------------
void cJVM::Inst_LAnd()
{
	jLong l1,l2;
	u4 iHighBytes,iLowBytes;
	//get first double
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	//assembly first double
	this->GetLong(l2,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	this->GetLong(l1,iHighBytes,iLowBytes);

	//add
	jLong l3 = l1 & l2;
	
	//disassembly result
	this->GetBytesFromLong(l3,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LCmp()
//-----------------------------------------------------------------------------
void cJVM::Inst_LCmp()
{
	jLong l1,l2;
	u4 iHighBytes,iLowBytes;
	//get first double
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	//assembly first double
	this->GetLong(l2,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	this->GetLong(l1,iHighBytes,iLowBytes);

	if(l1 > l2)
	{
		m_Stack.PushOperandInt(1);
	}
	else if(l1 == l2)
	{
		m_Stack.PushOperandInt(0);
	}
	else
	{
		m_Stack.PushOperandInt(-1);
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DAdd()
//-----------------------------------------------------------------------------
void cJVM::Inst_DAdd()
{
	jDouble d1,d2;
	u4 iHighBytes,iLowBytes;
	//get first jDouble
	iLowBytes = (u4)m_Stack.PopOperand().iU4;
	iHighBytes = (u4)m_Stack.PopOperand().iU4;

	//assembly first jDouble
	this->GetDouble(d2,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	this->GetDouble(d1,iHighBytes,iLowBytes);

	//add
	jDouble d3 = d1 + d2;
	
	//disassembly result
	this->GetBytesFromDouble(d3,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DSub
//-----------------------------------------------------------------------------
void cJVM::Inst_DSub()
{
	jDouble d1,d2;
	u4 iHighBytes,iLowBytes;
	//get first jDouble
	iLowBytes = (u4)m_Stack.PopOperand().iU4;
	iHighBytes = (u4)m_Stack.PopOperand().iU4;

	//assembly first jDouble
	this->GetDouble(d2,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	this->GetDouble(d1,iHighBytes,iLowBytes);

	//add
	jDouble d3 = d1 - d2;
	
	//disassembly result
	this->GetBytesFromDouble(d3,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DRem()
//-----------------------------------------------------------------------------
void cJVM::Inst_DRem()
{
	jDouble d1,d2;
	u4 iHighBytes,iLowBytes;
	//get first jDouble
	iLowBytes = (u4)m_Stack.PopOperand().iU4;
	iHighBytes = (u4)m_Stack.PopOperand().iU4;

	//assembly first jDouble
	this->GetDouble(d2,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	this->GetDouble(d1,iHighBytes,iLowBytes);

	//add
	jDouble d3 = fmod(d1,d2);
	
	//disassembly result
	this->GetBytesFromDouble(d3,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DNeg()
//-----------------------------------------------------------------------------
void cJVM::Inst_DNeg()
{
	jDouble d;
	u4 iHighBytes,iLowBytes;
	//get first jDouble
	iLowBytes = (u4)m_Stack.PopOperand().iU4;
	iHighBytes = (u4)m_Stack.PopOperand().iU4;

	//assembly first jDouble
	this->GetDouble(d,iHighBytes,iLowBytes);

	d = -d;
	
	//disassembly result
	this->GetBytesFromDouble(d,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DMul()
//-----------------------------------------------------------------------------
void cJVM::Inst_DMul()
{
	jDouble d1,d2;
	u4 iHighBytes,iLowBytes;
	//get first jDouble
	iLowBytes = (u4)m_Stack.PopOperand().iU4;
	iHighBytes = (u4)m_Stack.PopOperand().iU4;

	//assembly first jDouble
	this->GetDouble(d2,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	this->GetDouble(d1,iHighBytes,iLowBytes);

	//add
	jDouble d3 = d1 * d2;
	
	//disassembly result
	this->GetBytesFromDouble(d3,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DDiv()
//-----------------------------------------------------------------------------
void cJVM::Inst_DDiv()
{
	jDouble d1,d2;
	u4 iHighBytes,iLowBytes;
	//get first jDouble
	iLowBytes = (u4)m_Stack.PopOperand().iU4;
	iHighBytes = (u4)m_Stack.PopOperand().iU4;

	//assembly first jDouble
	this->GetDouble(d2,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iU4;
	iHighBytes = m_Stack.PopOperand().iU4;

	this->GetDouble(d1,iHighBytes,iLowBytes);

	//add
	jDouble d3 = d1 / d2;
	
	//disassembly result
	this->GetBytesFromDouble(d3,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LConst_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_LConst_0()
{
	jLong l = 0;
	u4 iHighBytes,iLowBytes;
	//disassembly result
	this->GetBytesFromLong(l,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LConst_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_LConst_1()
{
	jLong l = 1;
	u4 iHighBytes,iLowBytes;
	//disassembly result
	this->GetBytesFromLong(l,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DConst_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_DConst_0()
{
	jDouble d = 0;
	u4 iHighBytes,iLowBytes;
	//disassembly result
	this->GetBytesFromDouble(d,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DConst_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_DConst_1()
{
	jDouble d = 1;
	u4 iHighBytes,iLowBytes;
	//disassembly result
	this->GetBytesFromDouble(d,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_D2I()
//-----------------------------------------------------------------------------
void cJVM::Inst_D2I()
{
	jDouble d;
	u4 iHighBytes,iLowBytes;

	//get and assambly double
	iLowBytes = (u4)m_Stack.PopOperand().iU4;
	iHighBytes = (u4)m_Stack.PopOperand().iU4;

	this->GetDouble(d,iHighBytes,iLowBytes);

	//convert to integer
	jInt i = (jInt)d;
	//push result
	m_Stack.PushOperandInt(i);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_D2F()
//-----------------------------------------------------------------------------
void cJVM::Inst_D2F()
{
	jDouble d;
	u4 iHighBytes,iLowBytes;

	//get and assambly double
	iLowBytes = (u4)m_Stack.PopOperand().iU4;
	iHighBytes = (u4)m_Stack.PopOperand().iU4;

	this->GetDouble(d,iHighBytes,iLowBytes);

	//convert to integer
	jFloat f = (jFloat)d;
	//push result
	m_Stack.PushOperandFloat(f);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_D2L()
//-----------------------------------------------------------------------------
void cJVM::Inst_D2L()
{
	jDouble d;
	u4 iHighBytes,iLowBytes;

	//get and assambly double
	iLowBytes = (u4)m_Stack.PopOperand().iU4;
	iHighBytes = (u4)m_Stack.PopOperand().iU4;

	this->GetDouble(d,iHighBytes,iLowBytes);

	//convert to long
	jLong l = (jLong)d;

	this->GetBytesFromLong(l,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_L2D()
//-----------------------------------------------------------------------------
void cJVM::Inst_L2D()
{
	jLong l;
	u4 iHighBytes,iLowBytes;

	//get and assambly double
	iLowBytes = (u4)m_Stack.PopOperand().iU4;
	iHighBytes = (u4)m_Stack.PopOperand().iU4;

	this->GetLong(l,iHighBytes,iLowBytes);

	//convert to double
	jDouble d = (jDouble)l;

	this->GetBytesFromDouble(d,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_L2F()
//-----------------------------------------------------------------------------
void cJVM::Inst_L2F()
{
	jLong l;
	u4 iHighBytes,iLowBytes;

	//get and assambly double
	iLowBytes = (u4)m_Stack.PopOperand().iU4;
	iHighBytes = (u4)m_Stack.PopOperand().iU4;

	this->GetLong(l,iHighBytes,iLowBytes);

	//convert to double
	jFloat f = (jFloat)l;

	//push result
	m_Stack.PushOperandFloat(f);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_L2I()
//-----------------------------------------------------------------------------
void cJVM::Inst_L2I()
{
	jLong l;
	u4 iHighBytes,iLowBytes;

	//get and assambly double
	iLowBytes = (u4)m_Stack.PopOperand().iU4;
	iHighBytes = (u4)m_Stack.PopOperand().iU4;

	this->GetLong(l,iHighBytes,iLowBytes);

	//convert to double
	jInt i = (jInt)l;

	//push result
	m_Stack.PushOperandInt(i);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DCMPL()
//-----------------------------------------------------------------------------
void cJVM::Inst_DCMPL()
{
	//get value
	u4 iLow = (u4)m_Stack.PopOperand().iU4;
	u4 iHigh = (u4)m_Stack.PopOperand().iU4;
	jDouble value2;
	this->GetDouble(value2,iHigh,iLow);

	iLow = (u4)m_Stack.PopOperand().iU4;
	iHigh = (u4)m_Stack.PopOperand().iU4;
	jDouble value1;
	this->GetDouble(value1,iHigh,iLow);

	if(value1 > value2)
		m_Stack.PushOperandInt(1);
	else if(value1 < value2)
		m_Stack.PushOperandInt(-1);
	else if(value1 == value2)
		m_Stack.PushOperandInt(0);
	else
		m_Stack.PushOperandInt(-1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DCMPG()
//-----------------------------------------------------------------------------
void cJVM::Inst_DCMPG()
{
	//get value
	u4 iLow = (u4)m_Stack.PopOperand().iU4;
	u4 iHigh = (u4)m_Stack.PopOperand().iU4;
	jDouble value2;
	this->GetDouble(value2,iHigh,iLow);

	iLow = (u4)m_Stack.PopOperand().iU4;
	iHigh = (u4)m_Stack.PopOperand().iU4;
	jDouble value1;
	this->GetDouble(value1,iHigh,iLow);

	if(value1 > value2)
		m_Stack.PushOperandInt(1);
	else if(value1 < value2)
		m_Stack.PushOperandInt(-1);
	else if(value1 == value2)
		m_Stack.PushOperandInt(0);
	else
		m_Stack.PushOperandInt(1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FCMPL()
//-----------------------------------------------------------------------------
void cJVM::Inst_FCMPL()
{
	//get value
	jFloat value1, value2;
	value2 = m_Stack.PopOperand().fFloat;
	value1 = m_Stack.PopOperand().fFloat;

	if(value1 > value2)
		m_Stack.PushOperandInt(1);
	else if(value1 < value2)
		m_Stack.PushOperandInt(-1);
	else if(value1 == value2)
		m_Stack.PushOperandInt(0);
	else
		m_Stack.PushOperandInt(-1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_FCMPG()
//-----------------------------------------------------------------------------
void cJVM::Inst_FCMPG()
{
	//get value
	jFloat value1, value2;
	value2 = m_Stack.PopOperand().fFloat;
	value1 = m_Stack.PopOperand().fFloat;

	if(value1 > value2)
		m_Stack.PushOperandInt(1);
	else if(value1 < value2)
		m_Stack.PushOperandInt(-1);
	else if(value1 == value2)
		m_Stack.PushOperandInt(0);
	else
		m_Stack.PushOperandInt(1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_F2D()
//-----------------------------------------------------------------------------
void cJVM::Inst_F2D()
{
	jFloat f = m_Stack.PopOperand().fFloat;
	jDouble d = (jDouble)f;

	u4 iHighBytes,iLowBytes;
	//disassembly result
	this->GetBytesFromDouble(d,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_F2L()
//-----------------------------------------------------------------------------
void cJVM::Inst_F2L()
{
	jFloat f = m_Stack.PopOperand().fFloat;
	jLong l = (jLong)f;

	u4 iHighBytes,iLowBytes;
	//disassembly result
	this->GetBytesFromLong(l,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandU4(iHighBytes);
	m_Stack.PushOperandU4(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_F2I()
//-----------------------------------------------------------------------------
void cJVM::Inst_F2I()
{
	jFloat f = m_Stack.PopOperand().fFloat;
	jInt i = (jInt)f;

	//push result
	m_Stack.PushOperandInt(i);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_AAStore()
//-----------------------------------------------------------------------------
void cJVM::Inst_AAStore()
{
	OBJECT *pObj = m_Stack.PopOperand().pObject;
	jInt iIndex = m_Stack.PopOperand().iInt;

	OBJECT *pArrayObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pArrayObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (jInt)pArrayObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//skip header
	OBJECT **ppArray = (OBJECT**)(pArrayObj + 1);
	ppArray[iIndex] = pObj;
}

#define VM_DEBUG
//-----------------------------------------------------------------------------
// Name: cJVM::RunGC()
//-----------------------------------------------------------------------------
void cJVM::RunGC()
{
	#ifdef VM_DEBUG
	printf("\n\nVM: GC Started, allocated bytes: %d.\n",m_Heap.GetNumAllocatedBytes());
	#endif

	m_Heap.SwapHeapspace();

	//check stack
	for(u4 i = 2;i < this->m_Stack.GetTop();++i)
	{
		OBJECT *pObj = m_Stack.GetObject(i);
		if(pObj == NULL)
			continue;

		m_Stack.SetObject(i,CollectObject(pObj));
	}

	//check static members
	for(ClassMapItor i = m_Classes.begin();i != m_Classes.end();++i)
	{ //for all classes
		cJavaClass *pClass = i->second;
		//get static data pointer
		u1 *pData = pClass->m_pStaticData;

		//for all static fields
		for(u4 j = 0;j < pClass->GetNumFields();++j)
		{
			FIELD_INFO *pField = pClass->GetFieldInfo(j);
			if(pField->bReference == false)
				continue;
			if(pField->iAccessFlags & ACC_STATIC)
			{
				//get field
				OBJECT *pObj;
				memcpy(&pObj,&pData[pField->iOffset],sizeof(OBJECT*));
				if(pObj == NULL)
					continue;

				pObj = CollectObject(pObj);
				//collect field and update with new address
				memcpy(&pData[pField->iOffset],&pObj,sizeof(OBJECT*));
			}
		}
	}

	#ifdef VM_DEBUG
	printf("VM: GC Finished, allocated bytes: %d.\n\n",m_Heap.GetNumAllocatedBytes());
	#endif
}

//-----------------------------------------------------------------------------
// Name: cJVM::CollectObject()
//-----------------------------------------------------------------------------
OBJECT* cJVM::CollectObject(OBJECT *pObj)
{
	if(pObj->iArraySize == 0) //already handled, only update reference if neccesary
	{
		OBJECT *pNewObj = (OBJECT*)pObj->iSize;
		return (OBJECT*)pNewObj;
	}
	else
	{
		//check array size
		if(pObj->iArraySize > 1)
		{ 
			//array
			if(pObj->pClass != NULL)
			{ 
				//array of references -> check all
				//skip object header
				u1 *pObjData = (u1*)(pObj + 1);
				//go through all
				for(u4 i = 0;i < pObj->iArraySize;++i)
				{
					OBJECT *pArrayObj;
					memcpy(&pArrayObj,pObjData,sizeof(OBJECT*));

					if(pArrayObj == NULL)
					{
						pObjData += sizeof(OBJECT*);
						continue;
					}

					pArrayObj = CollectObject(pArrayObj);
					pObjData += sizeof(OBJECT*);
				}
			}

			//relocate array itself
			return m_Heap.RelocateObject(pObj);

		}
		else
		{ 
			//single obj - now for the difficult part
			u1 *pObjData = (u1*)(pObj + 1);

			cJavaClass *pClass = pObj->pClass;
			//go through all fields including super classes
			while(pClass != NULL)
			{
				for(int i = 0;i < pClass->GetNumFields();++i)
				{
					FIELD_INFO *pField = pClass->GetFieldInfo(i);
					if(pField->bReference == false)
						continue;
					if(pField->iAccessFlags & ACC_STATIC)
						continue;

					//get obj
					OBJECT *pFieldObj;
					memcpy(&pFieldObj,&pObjData[pField->iOffset],sizeof(OBJECT*));
					if(pFieldObj == NULL)
						continue;

					//collect obj
					memcpy(&pObjData[pField->iOffset],CollectObject(pFieldObj),sizeof(OBJECT*));
				}

				//try super class
				pClass = pClass->GetSuperClass();
			}			

			//relocate object itself
			return m_Heap.RelocateObject(pObj);
		}
	}
}