#include <vector>

//-----------------------------------------------------------------------------
//  Name: struct STACK_ITEM
//! Represents heap allocated object header
//-----------------------------------------------------------------------------
typedef struct STACK_ITEM
{
	union
	{
		OBJECT	*pObject;
		jInt	iInt;
		jFloat	fFloat;
		void	*pPointer;
		u4		iU4;
	};

} STACK_ITEM;

//-----------------------------------------------------------------------------
// Name: class cStack
// Desc: Implements script's stack
//-----------------------------------------------------------------------------
class cStack
{
private:
	typedef std::vector<STACK_ITEM> StackVec;
	typedef std::vector<u1> FlagVec;

	//Vector of stack values
	StackVec	m_Values;
	//Stack values flags
	FlagVec		m_Flags;
	//Index to top of the stack
	u4		m_iTopIndex;
	//Top Frame index
	u4		m_iFrameIndex;
	//Frame's variables index
	u4		m_iVariablesIndex;
	//Frame's operand stack top index
	u4		m_iOperandStackTopIndex;

public:
	//Constructor
	cStack();
	//Destructor
	~cStack();

	//init(clear) stack
	void Clear();

	//Get top of the stack
	u4 GetTop();
	//Get top of the operand stack
	u4 GetOperandStackTop();
	//Resize stack
	void Resize(u4 Size);
	//Push method's frame on top of stack
	bool PushMethodFrame(METHOD_INFO *pInfo,u1 *pPC,u2 iNumArgs);
	//Pop method's frame from top of stack
	u1* PopMethodFrame(cJavaClass **ppCurClass);
	//Return current frame's variable with index
	STACK_ITEM* GetVariable(u4 index);
	//Set current frame's variable - object
	void SetVariableObject(OBJECT *pObj,u4 index);
	//Set current frame's variable - jFloat
	void SetVariableFloat(jFloat data,u4 index);
	//Set current frame's variable - jInt
	void SetVariableInt(jInt data,u4 index);
	//Set current frame's variable - jInt
	void SetVariableU4(u4 data,u4 index);
	//Increment current frame's variable - jInt
	void IncVariableInt(jInt data,u4 index);
	//Get variable and push it on top of operand stack
	void PushVariable(u4 index);
	//Get variable and push it on top of operand stack - two values (double, long)
	void PushVariableDouble(u4 index);
	//Push object on top of operand stack
	void PushOperandObject(OBJECT *pObj);
	//Push operand on top of operand stack
	void PushOperand(STACK_ITEM Operand,bool bReference);
	//Push jInt on top of operand stack
	void PushOperandInt(jInt i);
	//Push jInt on top of operand stack
	void PushOperandU4(u4 i);
	//Push jFloat on top of operand stack
	void PushOperandFloat(jFloat f);
	//Duplicate top operand on operand stack
	void DuplicateTopOperand();
	//Duplicate top operand on operand stack and insert two values down
	void DuplicateX1();
	//Duplicate top operand on operand stack and insert three values down
	void DuplicateX2();
	//Duplicate top two operands on operand stack
	void DuplicateTopOperand2();
	//Duplicate top two operands on operand stack and insert three values down
	void Duplicate2X1();
	//Duplicate top two operands on operand stack and insert four values down
	void Duplicate2X2();
	//Pop operand from top of stack
	STACK_ITEM PopOperand();
	//Get operand with specified index
	STACK_ITEM GetOperand(jInt index);

	//Swaps two top most operands
	void SwapTopOperands();


	//Used during GC. Checks if item at index is an object reference, if so,
	//pointer is returned. Otherwise NULL is returned
	OBJECT* GetObject(jInt index);
	//Used during GC. Sets object reference to specified stack item.
	void SetObject(jInt index,OBJECT *pObj);
};